from dataclasses import dataclass, field
import networkx as nx
import scipy
import pandas as pd
import numpy as np
from cyberlabrat.core.utils import parallel_process




def calculate_correlation(method, x, y):
    """Calculate the correlation and p-value based on the specified method."""
    if method == "pearson":
        result = scipy.stats.pearsonr(x, y)
        return result.statistic, result.pvalue
    elif method == "spearman":
        result = scipy.stats.spearmanr(x, y)
        return result.correlation, result.pvalue
    elif method == "kendall":
        result = scipy.stats.kendalltau(x, y)
        return result.correlation, result.pvalue
    else:
        raise ValueError(f"Unknown method: {method}")


def correlate(method, return_type):
    """Return a correlation function based on the specified method, p-value threshold, and return type."""

    def executor(x, y):
        correlation, pvalue = calculate_correlation(method, x, y)
        if return_type == "pvalues":
            return pvalue
        elif return_type == "correlations":
            return correlation
        else:
            raise ValueError(f"Unknown return type: {return_type}")

    return executor


@dataclass
class Matrix:

    """
    Creates a reusable matrix class for a given eperimnet. 
    Args:
        data (pd.DataFrame):    The original dataset.
        treatment (str):        Identifier for the treatment group.
        between (str):          Variable type for correlation (e.g., 'compound' or 'region').
        variables (str):         'var1-var2' to correlate from type 'between'. If only one: self correlation.
        accross (str): The column that will constitute the rows/cols of the matrix.
        columns (list[str]): Columns to include in the analysis. If None, all columns are included.
        n_minimum (int): Minumum occurnces of overlapping var1 and var2 to be correlated. Default = 5.
        method (str): Correlation method ('pearson', 'spearman', 'kendall'). Default = "pearson".
        pvalue_threshold (float): Threshold for significance in correlation. Defult = 0.05

    Returns:
        filtered_data (pd.DataFrame): Subselected data filtered based on n_minimum between vairables.
        pivot (pd.DataFrame): Pivot table of the filtered data.
        corr_masked (pd.DataFrame): Masked correlation matrix based on p-value threshold.
        correlations (pd.DataFrame): Full correlation matrix.
        pvalues (pd.DataFrame): Matrix of p-values for the correlations.
        missing_values (list): List of variables with missing data (< n_minimum).
        missing_overlap (list): List of variable pairs with insufficient data overlap.

    """

    data: pd.DataFrame
    grouping: str
    between: str
    var1: str
    var2: str
    accross: str
    order: list[str] = None
    n_minimum: int = 5
    method: str = "pearson"
    pvalue_threshold: float = 0.05
    delay_execution: bool = field(default=True, kw_only=True)

    filtered_data: pd.DataFrame = field(init=False)
    pivot: pd.DataFrame = field(init=False)
    corr_masked: pd.DataFrame = field(init=False)
    correlations: pd.DataFrame = field(init=False)
    pvalues: pd.DataFrame = field(init=False)
    missing_values: list = field(init=False)
    missing_overlap: list = field(init=False)
    
    def __call__(self):
        self.__post_init__()
        return self

    def __post_init__(self):
        if self.delay_execution:
            self.delay_execution = False
        else:    
            self.is_square = self.var1 != self.var2
            self.filter_missing_values()
            self.pivot_data()
            self.order_columns()
            self.correlate()
            self.find_missing_overlap()
            self.process_triangle_correlogram()
            self.get_title

    def get_title(self):
        """
        Generates a title for the correlogram based on the matrix configuration.

        Returns:
            str: A title string.
        """
        if self.is_square:
            return f"{'-'.join([self.var1, self.var2])} in {self.grouping}"
        return f"{self.var1} in {self.grouping}"  
    
    def filter_missing_values(self):
        """
        Filters out variables with occurrences less than n_minimum and updates missing_values list.
        """
        self.missing_values = []
        missing_indices = []
        for col, df in self.data.groupby(by=[self.between, self.accross]):
            if df.value.notna().sum() < self.n_minimum:
                self.missing_values.append(col)
                missing_indices.extend(df.index)
        self.filtered_data = self.data.drop(missing_indices)
        if self.missing_values:
            print(
                f"{self.grouping} missing data for {self.missing_values}, deleted from analysis"
            )

    def pivot_data(self):
        """
        Creates a pivot table from the filtered data.
        """
        self.pivot = self.filtered_data.pivot_table(
            values="value",
            index=self.filtered_data["mouse_id"],
            columns=[self.between, self.accross],
        )

    def order_columns(self):
        """
        Orders the columns of the pivot table based on the provided column list.
        """
        columns = (
            sorted(
                self.pivot.columns,
                key=lambda x: self.order.index(x[1])
                if x[1] in self.order
                else float("inf"),
            )
            if self.order
            else self.pivot.columns
        )
        self.pivot = self.pivot[columns]
        
    def correlate(self):
        """
        Calculates and stores correlation and p-value matrices.
        """
        self.pvalues = self.create_corr_matrix('pvalues')
        self.correlations = self.create_corr_matrix('correlations')
        self.corr_masked = self.correlations[self.pvalues < self.pvalue_threshold]
        

    def create_corr_matrix(self, result_type):
        """
        Creates a correlation matrix for either correlation values or p-values.

        Args:
            result_type (str): Type of result to return ('pvalues' or 'correlations').

        Returns:
            pd.DataFrame: A DataFrame containing the requested correlation matrix.
        """
        method = correlate(self.method, result_type)
        return self.pivot.corr(method=method, min_periods=self.n_minimum).loc[
            tuple([self.var1, self.var2])
        ]

    def find_missing_overlap(self):
        """
        Identifies and reports variable pairs with insufficient data overlap.
        """
        self.missing_overlap = [
            ((self.var1, index), (self.var2, column))
            for (index, column), is_na in self.correlations.T.isna().stack().items()
            if is_na
        ]
        if self.missing_overlap:
            print(
                f"{self.grouping} insuficient overlapp for {self.missing_overlap} pairs"
            )
            print("Inspect with self.corr to adjust {columns} and redo analysis")

    def process_triangle_correlogram(self):
        """
        Masks the upper triangle of the correlogram if the matrix is not square.
        """
        if not self.is_square:
            mask = np.triu(np.ones(self.corr_masked.shape, dtype=bool), k=1)
            self.corr_masked[mask] = np.nan
            np.fill_diagonal(self.corr_masked.values, 1)

@dataclass
class Network:
    """
    A class to represent a network/graph constructed from a correlation matrix.

    Attributes:
    matrix (Matrix): An instance of the Matrix class containing the data and correlation matrix.

    Methods:
    max_node_degree(): Returns the maximum degree of the nodes in the graph.
    is_directed(): Property that checks if the network is directed (based on the matrix being square).
    get_title(): Generates a title for the network graph.
    plot_ax(ax): Plots the network graph on the given matplotlib axis.
    """
    matrix: Matrix
    delay_execution: bool = field(default=True, kw_only=True)
    
    def __call__(self):
        self.__post_init__()
        return self

    def __post_init__(self):
        """
        Initializes the network/graph from the given matrix. Constructs a directed or undirected graph
        based on the matrix's properties and fills it with nodes and edges based on the correlation data.
        """
        if self.delay_execution:
            self.delay_execution = False
        else:
            self.is_directed = self.matrix.is_square
            self.G = nx.MultiDiGraph() if self.matrix.is_square else nx.Graph()
            # directed edge -  to_correlate[0] --> to_correlate[1]
            self.G.clear()

            self.G.add_nodes_from(
                self.matrix.corr_masked.columns.tolist()
            )  # adds every BR as a node
            self.edge_labels = {}
            for (row, col), correlation in self.matrix.corr_masked.stack().dropna().items():
                # Add edge to the graph with edge weight and color
                # Avoid self sorrelation
                if not(row == col and not self.is_directed):
                    self.G.add_edge(
                        row,
                        col,
                        weight=abs(correlation),
                        color="red" if correlation > 0 else "blue",
                    )
                    
                    self.edge_labels[(row, col)] = f"{correlation:.2f}"

            angles = np.linspace(
                0, 2 * np.pi, len(self.matrix.corr_masked.columns), endpoint=False
            )
            self.pos = {
                col: (np.cos(angles[i]), np.sin(angles[i]))
                for i, col in enumerate(self.matrix.corr_masked.columns)
            }

            self.total_edges, self.pos_edges, self.neg_edges = self.edge_count()
            self.density = self.calculate_graph_density()
            self.max_degree, self.average_degree = self.calculate_node_degree()
            self.avg_clust_coeff_unweighted, self.avg_clust_coeff_weighted = self.calculate_clustering_coefficient()

            # self.local_efficiency = self.calculate_local_efficiency()
            # self.global_efficiency = self.calculate_global_efficiency()
            # self.characteristic_path_length = self.calculate_characteristic_path_length()

    def get_title(self):
        """Generates a formatted title for the network graph."""
        title = self.matrix.get_title()
        return title.replace('-', '->') if self.is_directed else title
    
    def edge_count(self):
        """
        Returns the total number of edges, positive edges, and negative edges in the graph.

        Returns:
            total_edges (int): The total number of edges in the graph.
            pos_edges (int): The number of edges with positive weights.
            neg_edges (int): The number of edges with negative weights.
        """
        total_edges = self.G.number_of_edges()
        pos_edges = 0
        neg_edges = 0

        for u, v, data in self.G.edges(data=True):
            color = data.get('color')  
            if color == 'red':
                pos_edges += 1
            elif color == 'blue':
                neg_edges += 1
        return total_edges, pos_edges, neg_edges
    
    def calculate_node_degree(self):
        """
        Returns:
          max_degree(int): the maximum degree of the nodes in the graph.
        """
        # Calculate degrees for all nodes and find the maximum and mean
        degrees = dict(self.G.degree())
        max_degree = max(degrees.values())
        average_degree = np.mean(list(degrees.values()))
        return max_degree, average_degree


    def calculate_graph_density(self):
        """
        Returns:
           graph_density (float): The density of the graph; edges/all_possible_edges.
        """
        num_edges = self.G.number_of_edges()
        num_nodes = self.G.number_of_nodes()

        if self.is_directed: #directed graph have doubble possible edges
            max_edges = num_nodes * (num_nodes - 1)
        else:
            max_edges = num_nodes * (num_nodes - 1) / 2
        return num_edges / max_edges if max_edges > 0 else 0
    

    
    def calculate_clustering_coefficient(self):
        """
        Calculates the average unweighted and weighted clustering coefficients for the graph.

        Returns:
            avg_clust_coeff_unweighted (float): The average unweighted clustering coefficient of the graph.
            avg_clust_coeff_weighted (float): The average weighted clustering coefficient of the graph.
        """
        if self.is_directed:
            # For directed graphs, use nx.clustering with 'directed' and 'weight' parameters
            clust_coeff_unweighted = None #nx.clustering(self.G.to_undirected(), weight=None)  # Unweighted
            clust_coeff_weighted = None #nx.clustering(self.G.to_undirected(), weight='weight')  # Weighted
        else:
            # For undirected graphs, use nx.clustering without 'directed' parameter
            clust_coeff_unweighted = nx.clustering(self.G)  # Unweighted
            clust_coeff_weighted = nx.clustering(self.G, weight='weight')  # Weighted

        avg_clust_coeff_unweighted = sum(clust_coeff_unweighted.values()) / len(clust_coeff_unweighted) if clust_coeff_unweighted else 0
        avg_clust_coeff_weighted = sum(clust_coeff_weighted.values()) / len(clust_coeff_weighted) if clust_coeff_weighted else 0

        return avg_clust_coeff_unweighted, avg_clust_coeff_weighted
    
    # def calculate_local_efficiency(self):
    #     """
    #     Returns:
    #         avg_local_eff_unweighted (float):  average unweighted local efficiency for i nodes.
    #         avg_local_eff_weighted (float):  average weighted local efficiency for i nodes.
    #     """
    #     local_eff_unweighted = 0
    #     local_eff_weighted = 0
    #     total_nodes = len(self.G.nodes())

    #     for node in self.G.nodes():
    #         # Determine neighbors for directed and undirected graphs
    #         if self.is_directed:
    #             neighbors = list(set(nx.predecessors(self.G, node)) | set(nx.successors(self.G, node)))
    #         else:
    #             neighbors = list(nx.neighbors(self.G, node))

    #         if len(neighbors) > 1:
    #             subgraph = self.G.subgraph(neighbors)
    #             # Calculate unweighted local efficiency
    #             local_eff_unweighted += nx.global_efficiency(subgraph)

    #             # Calculate weighted local efficiency
    #             # Ensure weights are considered in the subgraph efficiency calculation
    #             local_eff_weighted += nx.global_efficiency(subgraph, weight='weight')

    #     # Calculate average efficiencies
    #     avg_local_eff_unweighted = local_eff_unweighted / total_nodes if total_nodes > 0 else 0
    #     avg_local_eff_weighted = local_eff_weighted / total_nodes if total_nodes > 0 else 0
    #     return avg_local_eff_unweighted, avg_local_eff_weighted

    # def calculate_global_efficiency(self):
    #     """
    #     Calculates the average unweighted and weighted global efficiency of the graph.

    #     Returns:
    #         avg_global_eff_unweighted (float): The average unweighted global efficiency of the graph.
    #         avg_global_eff_weighted (float): The average weighted global efficiency of the graph.
    #     """
    #     # Unweighted Global Efficiency
    #     avg_global_eff_unweighted = nx.global_efficiency(self.G)

    #     # Weighted Global Efficiency
    #     # Inverting weights for efficiency calculation as smaller weights imply stronger connections
    #     G_copy = self.G.copy()
    #     inverted_weights = {(u, v): 1 / data['weight'] for u, v, data in G_copy.edges(data=True)}
    #     nx.set_edge_attributes(G_copy, inverted_weights, 'inverted_weight')
    #     avg_global_eff_weighted = nx.global_efficiency(nx.stochastic_graph(G_copy, weight='inverted_weight'))


    #     return avg_global_eff_unweighted, avg_global_eff_weighted

    

    # def calculate_characteristic_path_length(self):
    #     """
    #     Calculates the average unweighted and weighted characteristic path length of the graph.

    #     Returns:
    #         avg_path_length_unweighted (float): The average unweighted characteristic path length of the graph.
    #         avg_path_length_weighted (float): The average weighted characteristic path length of the graph.
    #     """
    #     if nx.is_connected(self.G):
    #         avg_path_length_unweighted = nx.average_shortest_path_length(self.G)
            
    #         # Inverting weights for path length calculation as smaller weights imply stronger connections
    #         G_copy = self.G.copy()
    #         inverted_weights = {(u, v): 1 / data['weight'] for u, v, data in G_copy.edges(data=True)}
    #         nx.set_edge_attributes(G_copy, inverted_weights, 'inverted_weight')
    #         avg_path_length_weighted = nx.average_shortest_path_length(G_copy, weight='inverted_weight')
    #     else:
    #         avg_path_length_unweighted = None
    #         avg_path_length_weighted = None

    #     return avg_path_length_unweighted, avg_path_length_weighted
    


from dataclasses import dataclass, field
from typing import ClassVar
import os, platform, subprocess
from cyberlabrat.core.FileSystem import FileSystem
import re

def sanitize_filename(filename):
    illegal_chars = r'[<>:"/\\|?*]'
    sanitized = re.sub(illegal_chars, '_', filename)
    return sanitized


@dataclass
class Cacheable:
    """
    Base Abstract class for all Cacheable objects.
    The point of cacheables is to have runtime-like variables that reflect the state of what is stored to disk.
    Especially useful for configuration excel files such as Palette.
    Handles loading, saving and filesystem architecture (see FileSystem.py)
    Unless filepath is provided, it will be generated from filename and extension.
    If filepath is provided, it will be used as-is.
    All child classes must implement the generate(), load() and save() methods.
    Generate must either return something that can be saved or the save method should know how to handle it.
    
    Args:
        filepath (str, optional): The path to the file. Defaults to None.

    Returns:
        object: What the child class returns
    """

    extension: ClassVar[str] = None
    filename: ClassVar[str] = None
    filepath: str = field(default=None, kw_only=True)
    from_scratch: bool = field(default=False, kw_only=True)

    def __post_init__(self):
        """
        Generates the filepath from the filename and extension.
        If the Cacheable is not saved, it will be initialized.

        """
        if not self.filepath:
            if not self.filename:
                raise ValueError("Child classes must define filename")
            self.filename = sanitize_filename(self.filename)
            # Automatically extrat relevant params for laction building
            self.filepath = os.path.join(
                FileSystem.get_location(**self.__dict__), self.filename
            )
        if not self.extension:
            raise ValueError("Child classes must define extension") 
        # Remove extension if it has already been added
        filepath, _ = os.path.splitext(self.filepath)
        self.filepath = f"{filepath}.{self.extension}"
        if not self.is_saved or self.from_scratch:
            self.initialize()

    def generate(self):
        raise NotImplementedError(
            "This method should be implemented for all custom Cacheables"
        )

    def initialize(self):
        """
        Calls the generate method and saves what it return to disk.
        """
        data = self.generate()
        self.save(data) if data is not None else self.save()
        print(f"CREATED AND CACHED {self.filepath}")


    def load(self):
        raise NotImplementedError(
            "This method should be implemented for all custom Cacheables"
        )

    def save(self, data):
        raise NotImplementedError(
            "This method should be implemented for all custom Cacheables"
        )
        
    def delete(self):
        os.remove(self.filepath)
                
    def open(self):
        if self.is_saved:
            if platform.system() == "Windows":
                os.startfile(self.filepath)
            elif platform.system() == "Darwin":
                subprocess.call(("open", self.filepath))
            elif platform.system() == "Linux":
                print("Can't handle Linux")
            else:
                raise OSError("Unknown operating system")
        else:
            raise FileNotFoundError(self.filepath)

    @property
    def is_saved(self):
        return os.path.isfile(self.filepath)



import ipywidgets as widgets
from IPython.display import display, clear_output
from cyberlabrat.core.Metadata import (
    TreatmentInformation,
    ExperimentInformation,
)
from cyberlabrat.core.FileSystem import FileSystem
from cyberlabrat.core.HPLC import HPLC
from cyberlabrat.core.Figure import Correlogram, Histogram, Correlation, Network, Table
from cyberlabrat.core.DataSelection import DataSelection, QuantitativeDataSelection
from cyberlabrat.core.Constants import COMPOUNDS, COMPOUND_CLASSES, REGION_CLASSES, REGIONS


class OrderedSelectMultipleWithAll(widgets.SelectMultiple):
    def __init__(self, options, **kwargs):
        super().__init__(options=options, **kwargs)

        self.value_ordered = []

        self.observe(self.handle_change, names="value")

    def __setattr__(self, name: str, value: list) -> None:
        if name == "options":
            value = ("Select/Deselect all", *value)
        super().__setattr__(name, value)

    def __getattribute__(self, name: str) -> tuple:
        return (
            [item for item in super().__getattribute__(name) if item != "Select/Deselect all"]
            if name == "value"
            else super().__getattribute__(name)
        )

    def handle_change(self, change):
        new_values = set(change.new)
        old_values = set(change.old)

        if new_values == {"Select/Deselect all"}:
            if "Select/Deselect all" not in old_values:
                self.value = tuple(self.options)
            else:
                self.value = tuple()
        else:
            added = new_values - old_values
            new_value_ordered = [item for item in self.value_ordered if item in new_values]
            self.value_ordered = [*new_value_ordered, *added]

def get_data(**kwargs):
    kwargs.pop("from_scratch", None)
    display(DataSelection(**kwargs).data)


def get_stats(**kwargs):
    kwargs.pop("from_scratch", None)
    kwargs.pop("treatment", None)
    display(QuantitativeDataSelection(**kwargs).statistics_table)


figure_mapping = {
    "correlogram": Correlogram,
    "histogram": lambda **kwargs: Histogram(**{k: arg for k, arg in kwargs.items() if k != "treatment"}),
    "correlation": Correlation,
    "network": Network,
    "table": Table,
    "data": get_data,
    "stats": get_stats,
}
class_constants_mappings = {
    "compounds": COMPOUND_CLASSES,
    "regions": REGION_CLASSES,
}


def playground():
    projects = FileSystem().list_projects()
    project_dropdown = widgets.Dropdown(
        options=projects, value=projects[0], description="Project:"
    )
    figure_dropdown = widgets.Dropdown(
        options=figure_mapping.keys(),
        value=list(figure_mapping.keys())[0],
        description="Figure:",
    )
    region_ordered_select_multiple = OrderedSelectMultipleWithAll(
        description="Regions:",
        options=REGIONS.list,
        style={"description_width": "initial"},
    )
    region_class_checkbox = OrderedSelectMultipleWithAll(
        description="Classes:",
        options=REGION_CLASSES.list,
        style={"description_width": "initial"},
    )
    compound_ordered_select_multiple = OrderedSelectMultipleWithAll(
        description="Compounds:",
        options=COMPOUNDS.list,
        style={"description_width": "initial"},
    )
    compound_class_checkbox = OrderedSelectMultipleWithAll(
        description="Classes:",
        options=COMPOUND_CLASSES.list,
        style={"description_width": "initial"},
    )
    treatment_ordered_select_multiple = OrderedSelectMultipleWithAll(
        description="Treatments:", options=[], style={"description_width": "initial"}
    )
    experiment_dropdown = widgets.Dropdown(
        description="Experiment:", options=[], style={"description_width": "initial"}
    )
    outliers_dropdown = widgets.Dropdown(
        options=["calculated", "eliminated", False],
        value="calculated",
        description="Eliminate outliers:",
        style={"description_width": "initial"},
    )
    submit = widgets.Button(
        description="Submit", layout=widgets.Layout(width="auto", margin="10px")
    )
    
    def display_interface():
        display(figure_dropdown)
        print(
            "Press control + click to select/deselect multiple items, and shift + click for range select"
        )

        widgets_vbox = widgets.VBox(
            [
                widgets.HBox(
                    [
                        project_dropdown,
                        experiment_dropdown,
                        treatment_ordered_select_multiple,
                        outliers_dropdown,
                    ]
                ),
                widgets.HBox(
                    [
                        region_ordered_select_multiple,
                        region_class_checkbox,
                        compound_ordered_select_multiple,
                        compound_class_checkbox,
                    ]
                ),
                submit,
            ],
            layout=widgets.Layout(align_items="center", justify_content="space-around"),
        )
        display(widgets_vbox)

    display_interface()
    
    def on_item_multiselect_updated(type, class_multiselect):
        classes_constant_mapping = class_constants_mappings[type]
        def update_selected_classes(item_multiselect):
            selected_classes = []
            for selected_class in class_multiselect.value:
                if all(item in item_multiselect.new for item in classes_constant_mapping[selected_class]):
                    selected_classes.append(selected_class)
            class_multiselect.value = selected_classes 
        return update_selected_classes
    region_ordered_select_multiple.observe(on_item_multiselect_updated("regions", region_class_checkbox), names="value")
    compound_ordered_select_multiple.observe(on_item_multiselect_updated("compounds", compound_class_checkbox), names="value")
            

    def on_class_multiselect_updated(type, item_multiselect):
        classes_constant_mapping = class_constants_mappings[type]
        def add_class_items_to_item_multiselect(class_multiselect):
            items_to_add = []
            new_selected_classes = set(class_multiselect.new) - set(class_multiselect.old)
            for selected_class in new_selected_classes:
                for element in classes_constant_mapping[selected_class]:
                    items_to_add.append(element)
            added_items = [*item_multiselect.value, *items_to_add]
            removed_classes = set(class_multiselect.old) - set(class_multiselect.new)
            removed_items = []
            for removed_class in removed_classes:
                removed_items.extend(classes_constant_mapping[removed_class])
            item_multiselect.value = [item for item in added_items if item not in removed_items]
            
        return add_class_items_to_item_multiselect

    region_class_checkbox.observe(on_class_multiselect_updated("regions", region_ordered_select_multiple), names="value")
    compound_class_checkbox.observe(on_class_multiselect_updated("compounds", compound_ordered_select_multiple), names="value")

    def update_project(_):
        project = project_dropdown.value
        print(f"PROJECT: {project}")
        data = HPLC(project).df
        treatment_information = TreatmentInformation(project)
        experiment_information = ExperimentInformation(project)

        compounds_constant = COMPOUNDS.list
        data_compounds = data.compound.unique()
        regions_constant = REGIONS.list
        data_regions = data.region.unique()

        region_ordered_select_multiple.options = [
            region for region in regions_constant if region in data_regions
        ]
        compound_ordered_select_multiple.options = [
            compound for compound in compounds_constant if compound in data_compounds
        ] + [
            compound for compound in data_compounds if compound not in compounds_constant
        ]
        treatment_ordered_select_multiple.options = (
            treatment_information.label.unique()
        )
        experiment_dropdown.options = ["None"] + list(
            experiment_information.label.unique()
        )
        experiment_dropdown.value = "None"

        def on_experiment_dropdown_updated(experiment_dropdown):
            if experiment_dropdown.new != "None":
                treatment_ordered_select_multiple.value = (
                    experiment_information.df.select(
                        experiment=experiment_dropdown.new
                    ).treatments
                )
            else:
                treatment_ordered_select_multiple.value = ()

        experiment_dropdown.observe(on_experiment_dropdown_updated, names="value")

        def on_treatment_multiselect_updated(treatment_ordered_select_multiple):
            if set(treatment_ordered_select_multiple.new) not in [
                set(treatments) for treatments in experiment_information.df.treatments
            ]:
                experiment_dropdown.value = "None"

        treatment_ordered_select_multiple.observe(
            on_treatment_multiselect_updated, names="value"
        )

    project_dropdown.observe(update_project, names="value")
    
    def display_figure(_):
        clear_output(wait=True)
        display_interface()
        print(f"Generating {figure_dropdown.value}")
        figure_mapping[figure_dropdown.value](
            project=project_dropdown.value,
            compound=compound_ordered_select_multiple.value_ordered,
            region=region_ordered_select_multiple.value,
            experiment=None if experiment_dropdown.value == "None" else experiment_dropdown.value,
            treatment=treatment_ordered_select_multiple.value_ordered,
            remove_outliers=outliers_dropdown.value,
        )

    submit.on_click(display_figure)

    if len(projects) == 1:
        update_project(None)



from cyberlabrat.core.JSON import JSONMapping
from cyberlabrat.core.FileSystem import FileSystem
import os
from dataclasses import dataclass
import difflib
from cyberlabrat.core.questions import yes_or_no, select_one
import json


@dataclass
class ConstantRegistry(JSONMapping):
    """A JSON mapping for constants.
    Used for validation of predefined constants (regions, compounds, compound classes..)

    Args:
        filepath (str): The path to the JSON file.

    Returns:
        THe content of the JSON file as a dict.
    """

    filepath: str

    def generate(self):
        raise NotImplementedError(
            "Contant registries should be handled directly in the JSON"
        )

    def detect(self, key):
        """Uses difflib to detect the closest match in the list.

        Args:
            key (hasable): The key to detect.
            Used by get_valid_choice to detect the closest match.
            Used for incorrect constants (region name with mistake...)

        Returns:
            str: the closest match
        """
        case_converter = {k.upper(): k for k in self.list}
        lazy_guess_upper = difflib.get_close_matches(
            key.upper(), case_converter.keys(), n=1, cutoff=0.6
        )
        return case_converter[lazy_guess_upper[0]] if lazy_guess_upper else None

    def get_valid_choice(self, invalid_choice):
        """
        Tries to get a valid choice from the user.
        Uses difflib to detect the closest match in the list.
        If no match is found, it asks the user to select one from the list.

        Args:
            invalid_choice (_type_): Key that was invalid

        Returns:
            str: A valid choice from keys
        """
        lazy_guess = self.detect(invalid_choice)
        if lazy_guess:
            is_correct = yes_or_no(
                f"INVALID: {invalid_choice}. DETECTED {lazy_guess}: {self[lazy_guess]}. CONFIRM?"
            )
            if is_correct:
                return lazy_guess
        try:
            new_choice = select_one(
                f"INVALID: {invalid_choice}, SELECT FROM:", self.dict
            )
            while new_choice not in self:
                new_choice = select_one(
                    f"UNKNOWN CHOICE: {new_choice}, SELECT FROM:", self.dict
                )
            return new_choice
        except SystemExit:
            print("EDIT REGISTRY AND RETRY")
            exit(1)

    def order(self, iterable):
        return [item for item in self if item in iterable]


REGIONS = ConstantRegistry(filepath=os.path.join(FileSystem.CONSTANTS, "regions"))
COMPOUNDS = ConstantRegistry(filepath=os.path.join(FileSystem.CONSTANTS, "compounds"))
COMPOUND_CLASSES = ConstantRegistry(
    filepath=os.path.join(FileSystem.CONSTANTS, "compound_classes")
)
REGION_CLASSES = ConstantRegistry(
    filepath=os.path.join(FileSystem.CONSTANTS, "region_classes")
)
CIRCUITS = ConstantRegistry(filepath=os.path.join(FileSystem.CONSTANTS, "circuits"))
COMPOUNDS_AND_REGIONS = {"region": REGIONS, "compound": COMPOUNDS}



import json
from dataclasses import dataclass
from cyberlabrat.core.Cacheable import Cacheable
from typing import ClassVar
from cyberlabrat.core.utils import is_array_like


@dataclass
class JSONMapping(Cacheable):
    """Base class for JSON mappings.
    Abstract and currentty only used for constants.
    Used to manipulate JSON file as runtime dictionnaries while maintaing possibility to manually edit and have changes be reflected at runtime.
    """

    extension: ClassVar[str] = "json"

    def load(self) -> dict:
        with open(self.filepath) as outfile:
            mapping = json.load(outfile)
        return mapping

    def save(self, mapping):
        with open(self.filepath, "w") as json_file:
            json.dump(mapping, json_file)

    def add(self, key, value):
        mapping = self.load()
        mapping[key] = value
        self.save(mapping)

    def get(self, key, default=None):
        return self.dict.get(tuple(key) if is_array_like(key) else key, default)

    def values(self) -> list:
        return self.dict.values()

    def keys(self) -> list:
        return self.dict.keys()

    def items(self) -> list:
        return self.dict.items()

    def __contains__(self, key):
        if is_array_like(key):
            key = tuple(key)
        return key in self.list

    def __getitem__(self, key):
        return self.dict.get(key)

    def __setitem__(self, key, value):
        self.add(key, value)


    def __iter__(self):
        for item in self.list:
            yield item
        
    def __repr__(self) -> str:
        return "\n".join(f"{k}: {v}" for k, v in self.items())

    @property
    def reversed(self):
        return {
            tuple(value) if is_array_like(value) else value: key
            for key, value in self.items()
        }

    @property
    def dict(self) -> dict:
        return self.load()

    @property
    def list(self) -> list:
        return list(self.dict.keys())



from dataclasses import dataclass, field
from typing import get_args
import pandas as pd
from cyberlabrat.core.FileSystem import FileSystem
from cyberlabrat.core.Metadata import (
    ProjectInformation,
    TreatmentInformation,
    ExperimentInformation,
)
from cyberlabrat.core.HPLC import HPLC, Outliers
from cyberlabrat.core.Statistics import QuantitativeStatistic
from cyberlabrat.core.utils import is_array_like
import matplotlib.pyplot as plt
import seaborn as sns
from cyberlabrat.core.questions import input_list, yes_or_no
from cyberlabrat.core.Constants import REGION_CLASSES, COMPOUND_CLASSES


def convert_parameter_to_list(parameter):
    if parameter is None:
        raise ValueError("None should be handled specifically")
    if isinstance(parameter, str):
        return parameter.replace(", ", ",").replace(" ,", "").split(",")
    elif is_array_like(parameter):
        return list(parameter)
    return [parameter]


def whisker_plot(data, grouping):
    swarm_palette = {
        "normal": "green",
        "eliminated": "red",
        "kept": "blue",
    }

    EXTRA_COLORS = [
        "red",
        "orange",
        "yellow",
        "pink",
        "purple",
        "brown",
    ]

    for hue in data.palette_hue.unique():
        if "suspected" in hue:
            swarm_palette[hue] = EXTRA_COLORS.pop(0)

    ax = sns.boxplot(y="value", data=data, dodge=False)
    sns.swarmplot(
        ax=ax,
        y="value",
        data=data,
        hue="palette_hue",
        size=5,
        palette=swarm_palette,
        dodge=False,
        edgecolor="k",
        linewidth=1,
    )
    ax.set_title(grouping)
    plt.legend(bbox_to_anchor=(1.05, 1), loc="upper left")
    plt.tight_layout()
    plt.show()


def handle_outlier_selection(title, data):
    finished = False
    while not finished:
        data["palette_hue"] = data.apply(
            lambda row: (
                f"suspected (mouse_id={row.mouse_id})"
                if row.outlier_status == "suspected"
                else row.outlier_status
            ),
            axis=1,
        )
        whisker_plot(data, title)
        eliminated = input_list(
            f"Select outliers for {title}: input mouse_ids to eliminated or write 'none'"
        )

        def label_eliminated(row):
            if row.is_outlier:
                return "eliminated" if str(row.mouse_id) in eliminated else "kept"
            return row.outlier_status

        data.palette_hue = data.apply(label_eliminated, axis=1)
        whisker_plot(data, title)
        finished = yes_or_no("Confirm selection?")
    return data.palette_hue


@dataclass
class DataSelection:

    project: str | list | tuple = field(kw_only=True)
    experiment: str = field(kw_only=True, default=None)
    treatment: str = field(kw_only=True, default=None)
    compound: str | list | tuple = field(kw_only=True, default=None)
    region: str | list | tuple = field(kw_only=True, default=None)
    remove_outliers: str | bool = field(kw_only=True, default=None)
    data_source: str | bool = field(kw_only=True, default="hplc")
    p_value_threshold: float = field(kw_only=True, default=None)
    pool: str = field(kw_only=True, default=None)

    def __post_init__(self):
        if self.project not in FileSystem.list_projects():
            raise ValueError(f"Unknown project {self.project}")

        self.project_information = ProjectInformation(self.project)
        self.treatment_information = TreatmentInformation(self.project)
        self.experiment_information = ExperimentInformation(self.project)

        self.p_value_threshold = (
            self.p_value_threshold or self.project_information.p_value_threshold
        )

        self.data = HPLC(self.project).full_df
        self.compound = COMPOUND_CLASSES.get(self.compound, self.compound)
        self.region = REGION_CLASSES.get(self.region, self.region)

        self.experiment_options = self.experiment_information.experiments
        self.treatment_options = self.treatment_information.label.unique()
        self.compound_options = self.data.compound.unique()
        self.region_options = self.data.region.unique()
        self.remove_outliers_options = ["calculated", "eliminated", False]

        for name in ["experiment", "treatment", "compound", "region"]:
            options = getattr(self, name + "_options")
            if getattr(self, name) is not None:
                processed_parameter = self.process_parameter(name, options)
                setattr(self, name, processed_parameter)
            if is_array_like(getattr(self, name)) and len(getattr(self, name)) == 1:
                setattr(self, name, getattr(self, name)[0])

        self.data = self.data.select(compound=self.compound, region=self.region)

        if self.experiment:
            self.experiment_information = self.experiment_information.select(
                label=self.experiment
            )

            self.data = self.data.select(experiment=self.experiment)

        if self.treatment:
            self.data = self.data.select(treatment=self.treatment)
            self.treatments = (
                [self.treatment] if isinstance(self.treatment, str) else self.treatment
            )
        else:
            self.treatments = [
                label
                for label in self.treatment_information.label
                if label in self.data.treatment.unique()
            ]

        if self.remove_outliers == "eliminated":
            self.process_outliers()
            self.data = self.data.select(outlier_status=["normal", "kept"])
        elif self.remove_outliers == "calculated":
            self.data = self.data.select(
                is_outlier=lambda x: x != True
            )  # nan considered not outlier

        self.data = self.data.select(value="notna")
        if self.pool: 
            self.data[self.pool] = "all"
            

    def process_parameter(self, name, options):
        parameter = getattr(self, name)
        parameter_to_list = convert_parameter_to_list(parameter)
        unknown_params = set(parameter_to_list) - set(options)
        if unknown_params:
            raise ValueError(f"Invalid parameter(s) for {name}: {unknown_params}")
        return parameter_to_list

    def process_outliers(self):
        for (treatment, compound, region), data in self.data.groupby(
            ["treatment", "compound", "region"]
        ):
            if "suspected" in data.outlier_status.values:
                title = f"{compound} in {region} for {treatment}"
                data.outlier_status = handle_outlier_selection(title, data)
                Outliers(self.project).update(data)


@dataclass()
class QuantitativeDataSelection(DataSelection):

    statistics_pipeline: list = field(kw_only=True, default=None)

    def __post_init__(self):
        if self.experiment == "weights":
            self.compound = "weight"
            
        super().__post_init__()
        if self.experiment is not None and self.pool != "treatment":
            self.statistics, self.statistics_table = (
                QuantitativeStatistic.calculate_from_selection(
                    self.data,
                    (
                        [self.experiment_information]
                        if self.experiment == 1
                        else self.experiment_information
                    ),
                    self.p_value_threshold,
                )
            )
            if len(self.statistics) == 1:
                self.statistic = self.statistics[0]
        else:
            self.statistics, self.statistics_table, self.statistic = [], pd.DataFrame(), None



import pandas as pd
import numpy as np
import os
from cyberlabrat.core.Dataset import ExcelDataset, SelectableDataFrame
from cyberlabrat.core.questions import select_one
from dataclasses import dataclass, field
from typing import ClassVar
from distutils.util import strtobool # Deprecated 3.12 https://stackoverflow.com/questions/715417/converting-from-a-string-to-boolean-in-python


@dataclass(repr=False)
class _ProjectSettings(ExcelDataset):
    """Base class for project settings.
    Handles loading, saving, and editing of project settings (excel files)

    Returns:
        ExcelDataset: Dataset with project settings
    """

    project: str = field(default=None)
    _template: ClassVar[dict] = None
    _template_types: ClassVar[dict] = None
    
    def __post_init__(self):
        """
        Load data from template and set columns as attributes
        """
        super().__post_init__()
        for key in self._template:
            self.__setattr__(key, self.df[key])

    def generate(self):
        """
        Generate template dataframe

        """
        return pd.DataFrame(self._template)

    def make_user_edit_excel(self):
        """
        Open excel file in edit mode to make user edit the config
        """
        self.open()
        question = "Press any key and ENTER when done editing file"
        user_finished = False
        while not user_finished:
            input(question)
            try:
                self.validate(self.load())
                user_finished = True
            except SystemExit:
                self.delete()
                print("System interuption, deleting file")
            except:
                question = "Error reading file. Press any key and ENTER when done correcting file"
                user_finished = False

    def initialize(self):
        """
        Same as chacheable initialize but also makes user edit as setting are user editable
        """
        super().initialize()
        self.make_user_edit_excel()

    def load(self):
        """
        Load data from template and handles type conversions for merges with raw data.
        This process makes sure the template is both human and programatic friendly.

        Raises:
            ValueError: If a cell is not of the correct type (ie user input unusable data)

        Returns:
            SelectableDataFrame: Contains the project settings
        """
        # vehicle.independant_var == nan, problem for "var in independant_var" (nan not iterable)
        data = super().load().replace(np.nan, "")
        try:
            return self.convert_dtypes(data)
        except:
            raise ValueError(f"Wrong data types, please correct {self.filename}")

    def convert_dtypes(self, df):
        for col_name, col_info in self._template_types.items():
            if col_info["type"] == list:
                df[col_name] = df[col_name].apply(
                    lambda val: [
                        col_info["subtype"](subval)
                        for subval in val.replace(" ", "").split(",")
                    ]
                )
            elif col_info["type"] == bool:
                df[col_name] = df[col_name].apply(lambda val: bool(strtobool(str(val))))
            else:
                df[col_name] = df[col_name].apply(col_info["type"])
        return df
    
    
    
    def __contains__(self, label):
        return label in self.df.label        


    def __getitem__(self, label) -> pd.Series:
        return self.df.select(**{"label": label})
    


@dataclass(repr=False)
class TreatmentInformation(_ProjectSettings):  # TODO: generalize to GroupInformation?

    filename: ClassVar[str] = "treatment_information"
    _template: ClassVar[dict] = {
        "group_id": [1, 5, 3, 4],
        "label": ["vehicles", "MDL", "TCB2", "TCB2+MDL"],
        "independant_variables": ["", "MDL", "TCB2", "TCB2, MDL"],
    }
    _template_types: ClassVar[dict] = {
        "group_id": {"type": int},
        "label": {"type": str},
        "independant_variables": {"type": list, "subtype": str},
    }
    control_group: ClassVar[list] = "vehicles"

    @property
    def palette(self):
        return {t.label: t.color for t in self}

    def get_hue_order(self):
        return self.df.sort_values(by="group_id").treatment.tolist()
    
    def load(self):
        data = super().load()
        data["treatment"] = data.label
        return data
    
    @property
    def treatments(self):
        return list(self.df.label)
    
    
@dataclass(repr=False)
class Palette(_ProjectSettings):  # TODO: generalize to GroupInformation?

    filename: ClassVar[str] = "palette"
    _template: ClassVar[dict] = {
        "treatment": ["vehicles", "MDL", "TCB2", "TCB2+MDL"],
        "color": ["white", "pink", "orange", "red"],
        "significance": ["*", "", "$", ""],
    }
    _template_types: ClassVar[dict] = {
        "treatment": {"type": str},
        "color": {"type": str},
        "significance": {"type": str},
    }

    @property
    def dict(self):
        return {t.treatment: t.color for t in self}
    
    
    def __contains__(self, value):
        return value in self.df.treatment
    
    def __getitem__(self, treatment) -> pd.Series:
        return self.df.select(**{"treatment": treatment})
    


@dataclass(repr=False)
class ExperimentInformation(_ProjectSettings):

    filename: ClassVar[str] = "experiment_information"
    _template: ClassVar[dict] = {
        "label": ["agonist antagonist"],
        "groups": ["1, 5, 3, 4"],
        "independant_variables": ["TCB2, MDL"],
        "paired": [False],
        "parametric": [True],        
        "control_group_id": [1]
    }
    _template_types: ClassVar[dict] = {
        "label": {"type": str},
        "groups": {"type": list, "subtype": int},
        "independant_variables": {"type": list, "subtype": str},
        "paired": {"type": bool},
        "parametric": {"type": bool},
        "control_group_id": {"type": int}
    }        

    def load(self):
        data = super().load()
        treatment_information = TreatmentInformation(self.project).df
        palette = Palette(self.project).dict
        full_experiment_info = []
        for _, experiment in data.iterrows():
            experiment["experiment"] = experiment.label
            experiment["treatments"] = treatment_information.select(group_id=experiment.groups).label.to_list()
            experiment["control_treatment"] = experiment.treatments[experiment.groups.index(experiment["control_group_id"])]
            experiment["palette"] = {t: palette[t] for t in experiment.treatments}
            full_experiment_info.append(experiment)
        return SelectableDataFrame(full_experiment_info)
    
    
    @property
    def experiments(self):
        return list(self.df.label)

def is_valid_file(file_path):
    if not os.path.isfile(file_path):
        print("Not found", file_path)
        return False
    extension = os.path.splitext(file_path)[1].lower()
    if extension not in [".xlsx", ".csv"]:
        print("Invalid extension:", extension)
        return False

    return True

@dataclass(repr=False)
class ProjectInformation(_ProjectSettings):

    filename: ClassVar[str] = "project_information"
    _template: ClassVar[dict] = {
        "label": ["TCB2"],
        "outlier_test": ["grubbs"],
        "p_value_threshold": [0.05],
        "raw_data_filename": ["raw_data.csv"],
    }
    _template_types: ClassVar[dict] = {
        "label": {"type": str},
        "outlier_test": {"type": str},
        "p_value_threshold": {"type": float},
        "raw_data_filename": {"type": str},
    }
    
    def generate(self):
        from cyberlabrat.core.HPLC import OUTLIER_TESTS
        data = super().generate()
        data["label"] = input("Enter project name")
        data["outlier_test"] = select_one("Select outlier test", OUTLIER_TESTS.keys())
        data["p_value_threshold"] = float(input("Enter p value threshold"))
        data["raw_data_filename"] = self.get_valid_filename()
        return data

    def get_valid_filename(self):
        # raw_data_filename = easygui.fileopenbox(title="Select raw HPLC file", filetypes=["*.xls", "*.xlsx"])
        raw_data_filename = input("Enter HPLC excel filename (must be in phd/)")
        file_path = f"{os.getcwd()}/{raw_data_filename}"

        while not is_valid_file(file_path):
            print(raw_data_filename, "NOT FOUND")
            raw_data_filename = input("Enter excel HPLC filename (must be in phd/)")
            file_path = f"{os.getcwd()}/{raw_data_filename}"

        return file_path

    @property
    def df(self):
        data = super().df
        return data.iloc[0] if len(data) == 1 else data

    def _repr_html_(self) -> str:
        return f"""
            <b>Project Information:</b><br>
            <ul>
                <li>Project: {self.project}</li>
                <li>Raw Data Filename: {self.raw_data_filename}</li>
                <li>Outlier Test: {self.outlier_test}</li>
                <li>P-Value Threshold: {self.p_value_threshold}</li>
            </ul>
        """


@dataclass(repr=False)
class DatasetInformation(_ProjectSettings):

    filename: ClassVar[str] = "dataset_information"
    _template: ClassVar[dict] = {
        "label": ["hplc", "tissue_weight"],
        "grouping_variables": ["treatment, compound, region", "treatment, region"],
        "independant_variables": ["TCB2, MDL", ""],
        "raw_data_filename": ["raw_data.csv"],
    }
    _template_types: ClassVar[dict] = {
        "label": {"type": str},
        "outlier_test": {"type": str},
        "p_value_threshold": {"type": float},
        "raw_data_filename": {"type": str},
    }


from __future__ import annotations
from dataclasses import dataclass
from cyberlabrat.core.Dataset import SelectableDataFrame
from cyberlabrat.core.questions import yes_or_no
from cyberlabrat.core.Metadata import (
    ExperimentInformation,
    TreatmentInformation,
    ProjectInformation,
)
from cyberlabrat.core.HPLC import RawHPLC, HPLC, Outliers
from cyberlabrat.core.Statistics import Statistics
import os

ROOT = f"{os.getcwd()}/PROJECTS" 
if not os.path.exists(ROOT):
        os.mkdir(ROOT)
@dataclass
class Project:
    """
    A class to manage and organize different components of a research project,
    including handling of experimental setups, data collection, and analysis.

    Attributes:
        name (str): The name of the project.
        location (str): The file path to the project's directory.
        project_information (ProjectInformation): Contains metadata about the project.
        experiment_information (ExperimentInformation): Handles the collection of experiments.
        experiments (dict): Stores instantiated Experiment objects, keyed by experiment names.
        treatment_information (TreatmentInformation): Manages information related to treatments in experiments.
        treatments (list): A list of treatments derived from treatment_information.
        raw_data (RawHPLC): An object handling raw high-performance liquid chromatography data.
        hplc (HPLC): An object to manage processed HPLC data.
        outliers (Outliers): Manages outlier detection based on project settings.
        statistics (Statistics): Handles statistical analysis of experimental data.
        full_df: the df with all values + outliers+ treatment nfo

    Methods:
        __post_init__(): Initializes a new Project instance, sets up necessary directories,
                         loads or prompts for project information, and prepares all related
                         components like experiments, treatments, and statistical analyses.
    """

    name: str
    
    def __post_init__(self):
        self.location = f"{ROOT}/{self.name}"
        if not os.path.exists(self.location):
            if yes_or_no(f"INITIALIZE NEW PROJECT: '{self.name}' ?"):
                os.mkdir(self.location)
            else:
                print(f"UNKNOWN PROJECT: {self.name}")
                print(f"KNOW PROJECTS ARE: {Project.list()}")
                exit(1)
        self.project_information = ProjectInformation(self.name)
        self.experiment_information = ExperimentInformation(self.name)
        self.treatment_information = TreatmentInformation(self.name)
        self.raw_data = RawHPLC(self.name)
        self.hplc = HPLC(self.name)
        self.outliers = Outliers(self.name)
        self.statistics = Statistics(self.name)        

    @property
    def data(self) -> SelectableDataFrame:
        return self.hplc.extend(self.treatment_information).extend(self.outliers)

    @classmethod
    def list(self):
        os.listdir(path=ROOT)


    def _repr_html_(self) -> str:
        return f"""
            <h3>Project: {self.name}</h3>
            <h4>Project Information</h4>
            {self.project_information._repr_html_()}
            <h4>Experiments</h4>
            {self.experiment_information._repr_html_()}
            <h4>Treatments</h4>
            {self.treatment_information._repr_html_()}
        """
    


import sys

def input_escape(question):
    answer = input(f"{question}\n(press ESCAPE to exit)")
    if answer == "":
        if yes_or_no("EXIT PROGRAM?"):
            sys.exit("PROGRAM TERMINATED BY USER")
    return answer

def select_one(question, options):
    return_key = True
    if not isinstance(options, dict):
        return_key = False
        options = {str(i+1): option for i, option in enumerate(options)}
    pretty_choices = ' -- '.join(': '.join(item) for item in sorted(options.items()))
    choice = input_escape(
        f"""{question}
        {pretty_choices}""")
    while choice not in options:
        choice = input_escape(f"""Invalid choice, possibilities are:\{pretty_choices}\n""")
    return choice if return_key else options[choice]

def yes_or_no(question):
    answer = input_escape(f"""{question} (y/n)\n""").upper()
    while answer not in ["Y", "N"]:
        answer = input_escape(f"""{question} (y/n)\n""").upper()
    return answer == "Y"


def input_list(question):
    return input_escape(f"""{question}\n (input values separated by comma)""").replace(" ", "").split(",")
    


from .FileSystem import FileSystem
from .Dataset import ExcelDataset, PickleDataset
from .Project import Project
from .HPLC import HPLC
from .Statistics import Statistics, QuantitativeStatistic
from .Figure import Histogram
from .Metadata import ProjectInformation, ExperimentInformation, Palette, TreatmentInformation
from .Constants import COMPOUNDS, COMPOUND_CLASSES, REGIONS, REGION_CLASSES, CIRCUITS


__all__ = [
    "FileSystem",
    "ExcelDataset",
    "PickleDataset",
    "Project",
    "HPLC",
    "Statistics",
    "QuantitativeStatistic",
    "Histogram",
    "ProjectInformation",
    "ExperimentInformation",
    "Palette",
    "TreatmentInformation",
    "COMPOUNDS",
    "COMPOUND_CLASSES",
    "REGIONS",
    "REGION_CLASSES",
    "CIRCUITS",
]


from tqdm.contrib.concurrent import process_map
import itertools
from collections.abc import Iterable


def is_array_like(value):
    """Detect wheher a value is a non-string iterable

    Args:
        value (Any): Value to evaluate

    Returns:
        bool: True if value is a non-string iterable
    """
    return isinstance(value, Iterable) and not isinstance(value, str)


def call_case(case):
    """
    Calls the given case and returns its result.
    Used by parallel_process for complex calculation where passing case to a function is complex.

    Args:
        case (Callable): A callable object to be executed.

    Returns:
        Any: The result of the executed case.
    """
    return case()


def parallel_process(cases, executor=call_case, description="Processing"):
    """Executes the operations performed by {executor} once per case in cases in parralel

    Args:
        executor (function): function to be executed, must take a single argument and unpack if necessary
        cases (list(args)): different arguments to be used by executor. Case == [arg1, arg2, ...] for executor(case)

    Returns:
        List of results from executor
    """
    results = process_map(executor, cases, desc=description, chunksize=1)
    return results


def flatten(two_dimension_list):
    return list(itertools.chain.from_iterable(two_dimension_list))



from dataclasses import dataclass, field
from typing import ClassVar
import pandas as pd
import numpy as np
from cyberlabrat.core.Dataset import PickleDataset, SelectableDataFrame
from cyberlabrat.core.HPLC import HPLC
from cyberlabrat.core.Metadata import (
    ExperimentInformation,
    ProjectInformation,
)
from tqdm import tqdm
import scipy
import pingouin as pg
from statsmodels.stats.multicomp import pairwise_tukeyhsd
from cyberlabrat.core.utils import parallel_process
from IPython.display import HTML
from cyberlabrat.core.utils import is_array_like
import statsmodels.api as sm
from statsmodels.formula.api import ols



def get_quantitative_statistics_pipeline(
    multiple_factors, multiple_treatments, paired, parametric
):
    """
    Determines the statistical tests pipeline based on experiment design parameters.

    Args:
        multiple_factors (bool): Flag indicating if multiple factors are involved.
        multiple_treatments (bool): Flag indicating if multiple treatments are considered.
        paired (bool): Flag indicating if the design is paired.
        parametric (bool): Flag indicating if the tests should be parametric.

    Returns:
        list of str: A list of names representing the statistical tests to be applied based on the input parameters.
    """
    return {
        (False, False, False, True): ["ttest"],
        (False, False, True, True): ["paired_ttest"],
        (False, True, False, True): ["one_way_anova", "tukey"],
        (False, True, True, True): ["repeated_measures_anova", "paired_ttest"],
        (True, True, False, True): ["two_way_anova", "one_way_anova", "tukey"],
    }[(multiple_factors, multiple_treatments, paired, parametric)]


@dataclass
class QuantitativeStatistic:
    """
    Handles statistical analysis for quantitative data.
    Needs all the required data to perform the statistical analysis as
    well as parameters that define the statistics to be performed.
    TODO: This class is a messy and needs a lot of refactoring + finding a smarter way to do it.

    Raises:
        ValueError: _description_
        ValueError: _description_

    Returns:
        _type_: _description_
    """

    data: pd.DataFrame
    independant_variables: list[str]
    treatments: list[str]
    is_paired: bool
    is_parametric: bool
    p_value_threshold: float
    group_column: str = field(default="treatment", kw_only=True)
    delay_execution: bool = field(default=False, kw_only=True)
    metadata: dict = field(default_factory=dict, kw_only=True)
    pipeline: list = field(default=None, kw_only=True)

    def __post_init__(self):
        if self.delay_execution:
            self.delay_execution = False
        else:
            self.filtered_data = self.data.select(value="notna")
            # check all treatments present and enough data
            self.has_enough_data = set(
                self.filtered_data[self.group_column].unique()
            ) == set(self.treatments) and all(
                [
                    group_data.value.count() >= 5
                    for _, group_data in self.filtered_data.groupby(self.group_column)
                ]
            )
            self.pipeline = self.pipeline or get_quantitative_statistics_pipeline(
                len(self.independant_variables) >= 2,
                len(self.treatments) >= 2,
                self.is_paired,
                self.is_parametric,
            )
            self.statistical_test = self.pipeline[0]
            self.post_hoc_test = self.pipeline[-1]
            self.filtered_data = self.data.select(value="notna")
            if self.has_enough_data:
                self.results = SelectableDataFrame(self.execute_stats_pipeline())
                self.significant_pairs = (
                    self.results.select(test=self.post_hoc_test).p_value
                    if self.post_hoc_test in self.results.test.to_list()
                    else None
                )
            else:
                self.results = SelectableDataFrame(
                    [
                        {
                            "test": test,
                            "is_significant": False,
                            "result": "Not enough data",
                            "result_string": "n/a",
                            "p_value_threshold": self.p_value_threshold,
                            "p_value": np.nan,
                        } for test in self.pipeline + ["validation"]
                    ]
                )
            for key, val in self.metadata.items():
                self.results[key] = val
            self.is_significant = self.results.is_significant.all()

    @classmethod
    def calculate(
        cls,
        project: str,
        experiment: str = None,
        compound: str = None,
        region: str = None,
        p_value_threshold: float = None,
        remove_outliers="calculated",
    ):
        """
        Calculate statistical results for a given experiment, compound, and region.
        If one ore some of the parameters are None or lists, calculate the results for all possible combinations.

        Args:
            project (str): The name of the project.
            experiment (str|list): The name of the experiment.
            compound (str|list): The name of the compound.
            region (str|list): The name of the region.
            p_value_threshold (float, optional): The p-value threshold used for statistical analysis. Defaults to None.
            remove_outliers (str, optional): Whether to remove outliers. Must be 'eliminated', 'calculated', or False. Defaults to "calculated".

        Returns:
            SelectableDataFrame: Containing the statistical results.

        Raises:
            ValueError: If remove_outliers is not 'eliminated', 'calculated', or False.
        """

        data = HPLC(project).full_df

        if compound:
            compounds = (
                compound
                if is_array_like(compound)
                else compound.replace(" ", "").split(",")
            )
        else:
            compounds = data.compound.unique()

        if region:
            regions = (
                region if is_array_like(region) else region.replace(" ", "").split(",")
            )
        else:
            regions = data.region.unique()

        if experiment:
            experiment = (
                experiment
                if is_array_like(experiment)
                else experiment.replace(", ", "").split(",")
            )
            experiments = [
                ExperimentInformation(project).select(experiment=experiment)
                for experiment in experiment
            ]
        else:
            experiments = [experiment for experiment in ExperimentInformation(project)]

        p_value_threshold = (
            p_value_threshold or ProjectInformation(project).p_value_threshold
        )

        if remove_outliers == "eliminated":
            data = data.select(outlier_status=["normal", "kept"])
        elif remove_outliers == "calculated":
            data = data.select(is_outlier=lambda val: val != True)
        elif remove_outliers is not False:
            raise ValueError(
                "remove_outliers must be 'eliminated', 'calculated', or False"
            )

        data = data.select(
            region=regions,
            compound=compounds,
        )

        groupings = []

        for experiment in experiments:
            groupings.extend(
                [
                    cls(
                        data=group_data,
                        independant_variables=experiment.independant_variables,
                        treatments=experiment.treatments,
                        is_paired=experiment.paired,
                        is_parametric=experiment.parametric,
                        p_value_threshold=p_value_threshold,
                        delay_execution=True,
                        metadata={
                            "project": project,
                            "experiment": experiment.label,
                            "compound": compound,
                            "region": region,
                        },
                    )
                    for (region, compound), group_data in tqdm(
                        data.select(treatment=experiment.treatments).groupby(
                            ["region", "compound"]
                        ),
                        desc=f"Preparing statistical groupings for {experiment.label}",
                    )
                ]
            )

        statistics = parallel_process(groupings, description="Calculating statistics")

        results = []
        for statistic in statistics:
            result = statistic.results
            result["fully_significant"] = statistic.is_significant
            results.append(result)

        return SelectableDataFrame(pd.concat(results))
    
    @staticmethod
    def calculate_from_selection(
        data,
        experiments,
        p_value_threshold: float = None,
        pipeline: list = None,
    ):
        """
        Calculate statistical for data, autmaticcaly groups by experiment, compound regions.

        Args:
            data (pd.DataFrame): The name of the project.
            experiemnts (str): The experiments to calculate for.
            p_value_threshold (float, optional): The p-value threshold used for statistical analysis. Defaults to None.

        Returns:
            SelectableDataFrame: Containing the statistical results.

        """
        
        groupings = []
        
        experiments = [experiments] if isinstance(experiments, pd.Series) else experiments

        for experiment in experiments:
            groupings.extend(
                [
                    QuantitativeStatistic(
                        data=group_data,
                        independant_variables=experiment.independant_variables,
                        treatments=experiment.treatments,
                        is_paired=experiment.paired,
                        is_parametric=experiment.parametric,
                        p_value_threshold=p_value_threshold,
                        pipeline=pipeline,
                        delay_execution=True,
                        metadata={
                            "experiment": experiment.label,
                            "compound": compound,
                            "region": region,
                        },
                    )
                    for (region, compound), group_data in tqdm(
                        data.select(treatment=experiment.treatments).groupby(
                            ["region", "compound"]
                        ),
                        desc=f"Preparing statistical groupings for {experiment.label}",
                    )
                ]
            )

        statistics = parallel_process(groupings, description="Calculating statistics")

        results = []
        for statistic in statistics:
            result = statistic.results
            result["fully_significant"] = statistic.is_significant
            results.append(result)

        return statistics, SelectableDataFrame(pd.concat(results))

    # for parallel
    def __call__(self):
        self.__post_init__()
        return self 

    def execute_stats_pipeline(self):
        results = []
        for test in self.pipeline:
            test_method_name = f"get_{test}"
            if not hasattr(self, test_method_name):
                raise ValueError(f"Unknown test: {test}")
            test_method = self.__getattribute__(test_method_name)
            test_result = test_method()
            results.append(
                {
                    **test_result,
                    "test": test,
                    "p_value_threshold": self.p_value_threshold,
                }
            )
            self.__setattr__(test, test_result["result"])
        return results

    def get_tukey(self):
        """
        Performs Tukey's HSD (Honest Significant Difference) test to identify significant differences between groups.

        Args:
            data (pd.DataFrame): The DataFrame containing the data to analyze, which must include a 'value' column and a 'treatment' column.
            p_value_threshold (float): The significance level to determine if the results are statistically significant.

        Returns:
            tuple: A tuple containing:
                - A list of tuples, where each tuple contains the pairs of treatments that have a significant difference.
                - A boolean indicating if any significant results were found.
                - A DataFrame of the complete results from the Tukey HSD test.
        """
        columns, *stats_data = pairwise_tukeyhsd(
            endog=self.filtered_data.value,
            groups=self.filtered_data[self.group_column],
            alpha=self.p_value_threshold,
        )._results_table.data
        results = pd.DataFrame(stats_data, columns=columns)
        significance_infos = pd.DataFrame(
            list(
                results[results.reject].apply(
                    lambda res: [(res.group1, res.group2), res["p-adj"]], axis=1
                )
            ),
            columns=["pairs", "p_values"],
        )
        return {
            "p_value": [
                significance_infos.pairs.tolist(),
                significance_infos.p_values.tolist(),
            ],
            "is_significant": len(significance_infos) > 0,
            "result": results,
            "result_string": significance_infos,
        }

    def get_one_way_anova(self):
        """
        Performs a one-way ANOVA test to determine if there are any statistically significant differences between the means of three or more independent (unrelated) groups.

        Args:
            data (pd.DataFrame): The DataFrame containing the data, where 'value' is the dependent variable and 'treatment' is the independent variable used to define groups.
            p_value_threshold (float): The alpha level used to determine the threshold for significance.

        Returns:
            tuple: A tuple containing:
                - The p-value from the ANOVA test.
                - A boolean indicating if the test result is significant at the given threshold.
                - A DataFrame containing the ANOVA test results (F-value and p-value).
        """
        model = ols(f'value ~ C({self.group_column})', data=self.filtered_data).fit()
        anova_table = sm.stats.anova_lm(model, typ=2)
        p_value = anova_table["PR(>F)"][0]
        F = anova_table['F'][0]
        df1, df2 = anova_table['df'][0], anova_table['df'][1]
    
        return {
            "p_value": p_value,
            "is_significant": p_value <= self.p_value_threshold,
            "result": anova_table,
            "result_string": f"F({df1}, {df2}) = {F}",
        }

    def get_two_way_anova(self):
        """
        Performs a two-way ANOVA to evaluate the effect of two nominal predictor variables on a continuous outcome variable.

        Args:
            data (pd.DataFrame): The DataFrame containing the experimental data. 'value' is the dependent variable. The independent variables should be specified as columns following the 'experiment' column.

        Returns:
            tuple: A tuple containing:
                - The p-value of the interaction term in the ANOVA table.
                - A boolean indicating if the interaction effect is significant below the given p_value_threshold.
                - A DataFrame with detailed ANOVA results including F-values and p-values for each effect.
        """
        data = self.filtered_data.copy()
        data[self.independant_variables] = data.apply(
            lambda row: [
                variable in row.independant_variables
                for variable in self.independant_variables
            ],
            axis=1,
            result_type="expand",
        )
        results = pg.anova(
            data=data,
            dv="value",
            between=self.independant_variables,
            detailed=True,
        ).round(3)
        
        
        return {
            "p_value": results["p-unc"][2],
            "is_significant": isinstance(results["p-unc"][2], float)
            and results["p-unc"][2] < self.p_value_threshold,
            "result": results,
            "result_string": f"F({results['DF'][2]}, {results['DF'][3]}) = {results['F'][2]}"
        }


@dataclass(repr=False)
class Statistics(PickleDataset):
    """
    Contains all quantitative statistical resuts for hplc

    """

    project: str
    filename: ClassVar[str] = "statistics"

    def select(self, **kwargs):
        """
        Enable selection of specific significance pairs
        Args:
            significant_pair: tuple of significant pairs

        """
        searched_pair = kwargs.pop("significant_pair", None)
        data = self.df.select(**kwargs)
        if searched_pair:
            return data[
                data.p_value.apply(
                    lambda p_value: (
                        set(searched_pair)
                        in [set(significant_pair) for significant_pair in p_value[0]]
                        if is_array_like(p_value)
                        else False
                    )
                )
            ]
        return data

    def generate(self):
        return QuantitativeStatistic.calculate(self.project).select(
            fully_significant=True
        )

    @property
    def significant_results(self):
        """Return groupings where all tests are significant.

        Returns:
            pd.DataFrame: Significant results.
        """
        return SelectableDataFrame(
            pd.concat(
                [
                    results
                    for _, results in self.df.groupby(
                        ["experiment", "compound", "region"]
                    )
                    if results.is_significant.all()
                ]
            )
        )

    @property
    def insufficent_data(self):
        """Return groupings where one group has less than 5 data points.

        Returns:
            pd.DataFrame: Insufficient data results.
        """
        try:
            invalid_groupings = self.select(test="validation")
            return invalid_groupings[invalid_groupings.result == "Not enough data"]
        except ValueError as e:
            if "EMPTY SELECTION" in str(e):
                return "No data"

        return self.df[
            (self.df.test == "validation" & self.df.result == "Not enough data")
        ]


@dataclass
class AggregateStatistics(PickleDataset):
    """
    Calculates group level descriptive statistics.
    """

    project: str
    filename: ClassVar[str] = "aggregate_statistics"

    def generate(self):
        result_ls = []
        for (treatment, region, compound), groupby_df in tqdm(
            HPLC(self.project).full_df.select(value="notna", is_outlier=False).groupby(by=["treatment", "region", "compound"]),
            desc="Calculating aggregate statistics",
        ):
            
            if len(groupby_df) >= 3:
                F, p, = scipy.stats.shapiro(groupby_df["value"])
                is_parametric = p > 0.05            
            else:
                F, p, is_parametric = np.nan, np.nan, np.nan
            
            mean, std, sem, values = [
                groupby_df.value.mean(),
                groupby_df.value.std(),
                groupby_df.value.sem(),
                groupby_df.value.values,
            ]
            result_ls.append(
                [treatment, region, compound, F, p, is_parametric, mean, std, sem, values]
            )
        return pd.DataFrame(
            result_ls,
            columns=[
                "treatment",
                "region",
                "compound",
                "shapiro_F",
                "shapiro_p",
                "is_parametric",
                "mean",
                "std",
                "sem",
                "values",
            ],
        )
        
        



from cyberlabrat.core.Statistics import QuantitativeStatistic, AggregateStatistics
import unittest
from cyberlabrat.core.Project import Project, ROOT
from cyberlabrat.core.Figure import Histogram, Correlation, Correlogram, Network, Table, StatisticsTable
import os, shutil


class TestProject(unittest.TestCase):
    def setUp(self):
        # Setup initial conditions here if necessary
        self.project_name = "TEST"
        self.project_location = f"{ROOT}/{self.project_name}"
        self.raw_data_filename = "raw_data.xlsx"

        # If deletion idn't happen (quit debug session too early for ex)
        raw_data_location = os.path.join(os.getcwd(), self.raw_data_filename)
        if os.path.exists(raw_data_location):
            os.remove(raw_data_location)
        if os.path.exists(self.project_location):
            shutil.rmtree(self.project_location)

        os.mkdir(self.project_location)

        # Copy necessary test data to the project location
        test_data_dir = f"{os.getcwd()}/tests/test_project_data/"
        for filename in os.listdir(test_data_dir):
            test_data_file = os.path.join(test_data_dir, filename)
            project_test_location = os.path.join(self.project_location, filename)
            if filename == self.raw_data_filename:
                shutil.copy(test_data_file, os.getcwd())
            else:
                shutil.copy(test_data_file, project_test_location)

    def test_project(self):
        # This runs the test
        project = Project(self.project_name)
        assert project.data.columns.to_list() == [
            "mouse_id",
            "group_id",
            "value",
            "compound",
            "region",
            "treatment",
            "color",
            "independant_variables",
            "label",
            "is_outlier",
            "outlier_status",
        ]
        assert len(project.data) == 1152
        assert (
            len(
                project.data.select(is_outlier=True).select(
                    treatment=["vehicles", "MDL"]
                )
            )
            == 18
        )
        stats = QuantitativeStatistic.calculate(
            project=self.project_name,
            experiment="agonist antagonist",
            compound="DA/5HT",
            region="OF",
            p_value_threshold=0.05,
        )
        assert stats.is_significant.to_list() == [True, True, True]
        stats = QuantitativeStatistic.calculate(
            project=self.project_name,
            experiment="agonist antagonist",
            compound="DA/5HT",
            region="OF",
            p_value_threshold=0.01,
        )
        assert stats.is_significant.to_list() == [False, True, False]
        stats = QuantitativeStatistic.calculate(
            project="TCB2", experiment="agonist antagonist", compound="5HT"
        )
        AggregateStatistics(project=self.project_name)
        assert len(AggregateStatistics(project=self.project_name).select(is_parametric=True, compound=lambda compound: '/' not in compound)) == 20
        assert len(AggregateStatistics(project=self.project_name).select(is_parametric=False)) == 4
        assert len(stats.select(fully_significant=True)) == 9
        assert len(project.statistics.significant_results) == 6

        assert Histogram(
            project="TCB2",
            experiment="dose response",
            compound="5HIAA/5HT",
            region="OF",
            from_scratch=True,
        ).statistics_table.is_significant.to_list() == [True, True]
        assert Histogram(
            project="TCB2",
            experiment="dose response",
            compound="5HIAA/5HT",
            from_scratch=True,
            remove_outliers="calculated",
        ).statistics_table.is_significant.all()
        legit_regions = [
            "OF",
            "PL",
            "aCC",
            "M1",
            "SJ",
            "S1L",
            "S1R",
            "AC",
            "V1",
            "A",
            "dH",
            "vH",
            "NAc",
            "VM",
            "DM",
            "VL",
            "DL",
            "MD",
            "VPL",
            "VPR",
            "DLG",
            "HY",
            "SC",
            "SN",
            "VTA",
            "DR",
            "MR",
            "CB",
        ]
        Correlogram(
            project="TCB2",
            compound="5HT-DA",
            region=legit_regions,
            from_scratch=True,
            remove_outliers="calculated",
        )
        Correlation(
            project="TCB2",
            experiment="agonist antagonist",
            treatment="vehicles",
            compound="DA",
            region=["VPL", "SN"],
            from_scratch=True,
            remove_outliers="calculated",
        )
        node_positions = {
            "OF": (5.9, 2.4),
            "PL": (5.6, 3.5),
            "aCC": (7.1, 4),
            "M1": (6.2, 3.8),
            "SJ": (7.5, 3.8),
            "S1L": (8.5, 3.8),
            "S1R": (8.2, 4.1),
            "AC": (9.9, 4.2),
            "V1": (10.9, 4),
            "MD": (10.3, 1.45),
            "VPR": (9.1, 1.45),
            "VPL": (9.4, 1.15),
            "DLG": (9.7, 1.9),
        }

        Network(
            project="TCB2",
            node_positions=node_positions,
            compound="5HT",
            region="thalamocortical_interaction",
            from_scratch=True,
            remove_outliers="calculated",
        )
        Table(
            project="TCB2",
            compound=["5HT", "5HIAA", "DA", "DOPAC", "HVA", "3MT", "NA"],
            treatment="vehicles",
            region=legit_regions,
            from_scratch=True,
            remove_outliers="calculated",
        )
        Histogram(project='TEST', experiment="weight", region='OF', from_scratch=True, remove_outliers="calculated")
        StatisticsTable(project="TCB2", compound=["DA", "5HT"], experiment ="agonist antagonist", from_scratch=True)
        Histogram(project='TCB2', 
                experiment='agonist antagonist',
                compound=['DA', 'NA'], 
                region=['OF', 'PL'],
                from_scratch=True, 
                remove_outliers='calculated',
                pool="treatment"
                )
        
    def tearDown(self):
        # Cleanup code here
        print("DELETING TEST FILES")
        os.remove(os.path.join(os.getcwd(), self.raw_data_filename))
        shutil.rmtree(self.project_location)
        print("TEST FILES DELETED")


if __name__ == "__main__":
    unittest.main()



import os
import numpy as np
import pandas as pd
from dataclasses import dataclass, field
from typing import ClassVar
from cyberlabrat.core.Cacheable import Cacheable
import pandas as pd
from cyberlabrat.core.utils import is_array_like
ROOT = os.getcwd()  # This gives terminal location (terminal working dir)


def mask(df: pd.DataFrame, mask_conditions: dict):
    selected = df.index != None  # Select all
    absent_columns = set(mask_conditions) - set([*df.columns, 'index'])
    if absent_columns:
        raise ValueError(
            f"Unknown columns: {absent_columns}, possible columns are {df.columns}"
        )
    for key, value in mask_conditions.items():  # Refine selection
        column = pd.Series(df.index) if key == "index" else df[key] 
        if value is None:
            print(f"Skipping {column.name}, .select() ignores None for practical purpose s, use 'nan' (str) instead.")
        else:
            if callable(value):
                sub_selection = column.apply(value)
            elif is_array_like(value):
                sub_selection = column.isin(value)
            else:
                if value in ["nan", "notna"]:
                    sub_selection = column.isna() if value is None else column.notna()
                else:
                    sub_selection = column == value
            selected &= sub_selection
    return selected


def sub_select(df, selector, copy=True):
    df = df.loc[mask(df, selector)]
    return df.iloc[0] if len(df) == 1 else df


class SelectableDataFrame(pd.DataFrame):

    @property
    def _constructor(self):
        return SelectableDataFrame

    def select(self, **selector) -> "SelectableDataFrame":
        """
        Filter the DataFrame based on a selector.

        Args:
            selector (dict): A dictionary of column conditions to filter by. 
            'nan' and 'notna' are supported using strings. 
            None is ignored for dict unpacking purposes and because it is not a valid value.

        Returns:
            SelectableDataFrame: Filtered DataFrame that also includes the select method.
            Series: if selection conditions result in a single row
        """
        sub_selection = sub_select(self, selector)
        return sub_selection
    
    def extend(self, df: "Dataset|SelectableDataFrame|pd.DataFrame") -> "SelectableDataFrame":
        """
        Extend the DataFrame with another DataFrame. Automatically selects common columns.

        Args:
            df (_type_): the df to left join to self

        Returns:
            SelectableDataFrame:  Resulting DataFrame of left join
        """
        if isinstance(df, Dataset):
            df = df.df
        common_columns = self.columns.intersection(df.columns).to_list()
        return self.merge(df, on=common_columns, how="left")

@dataclass
class Dataset(Cacheable):
    """
    Base class for datasets ie dataframes stored in Excel or Pickle files.
    Similar to JSONmapping interface for json/dict.
    Actual dataframe is accessed through the df property and read directly from the file.

    Returns:
        Dataset: Wrapper for dataframes
    """

    def select(self, **selector) -> SelectableDataFrame:
        return self.df.select(**selector)

    @property
    def list(self):
        return list(self.df.to_dict(orient="index").values())

    @property
    def df(self) -> SelectableDataFrame:
        return SelectableDataFrame(self.load())

    def extend(self, other) -> SelectableDataFrame:
        """
        Extend the DataFrame with another DataFrame. Automatically selects common columns.

        Args:
            df (_type_): the df to left join to self

        Returns:
            SelectableDataFrame:  Resulting DataFrame of left join
        """
        return self.df.extend(other)

    def replace(self, column, mapping):
        data = self.df
        self.save(data.replace(mapping))
        
    def __contains__(self, column):
        return column in self.df

    def __repr__(self) -> str:
        """Called by terminal to display the dataframe (pretty)

        Returns:
            str: Pretty representation of the df
        """
        return repr(self.df)

    def _repr_html_(self) -> str:
        """Called by jupyter notebook to display the dataframe as html (pretty)

        Returns:
            str: Pretty representation of the df
        """
        return self.df._repr_html_()
    
    def __iter__(self):
        return iter([row for _, row in self.df.iterrows()])


@dataclass
class PickleDataset(Dataset):
    """
    Dataset wrapper for pickle files

    """

    extension: ClassVar[str] = "pkl"

    def save(self, data: pd.DataFrame):
        data.to_pickle(self.filepath)

    def load(self) -> SelectableDataFrame:
        return SelectableDataFrame(pd.read_pickle(self.filepath))


@dataclass
class ExcelDataset(Dataset):
    """
    Dataset wrapper for excel files

    """

    extension: ClassVar[str] = "xlsx"

    def save(self, data: pd.DataFrame):
        data.to_excel(self.filepath)

    def load(self) -> SelectableDataFrame:
        return SelectableDataFrame(pd.read_excel(self.filepath))



from dataclasses import dataclass, field

import networkx as nx
from scipy.stats import norm
from scipy import stats

from cyberlabrat.core.Cacheable import Cacheable
from cyberlabrat.core.Dataset import SelectableDataFrame, ExcelDataset
from cyberlabrat.core.Statistics import QuantitativeStatistic
from cyberlabrat.core.Metadata import (
    Palette,
)
from cyberlabrat.core.Matrix import Matrix
from cyberlabrat.core.Matrix import Network as NetworkModel
from cyberlabrat.core.Constants import COMPOUNDS_AND_REGIONS, REGIONS
from matplotlib import pyplot as plt
import seaborn as sns
from typing import ClassVar
import numpy as np
from IPython.display import Image, display
import pandas as pd
import os
from cyberlabrat.core.utils import parallel_process
from statannotations.Annotator import Annotator
from cyberlabrat.core.DataSelection import DataSelection, QuantitativeDataSelection
from cyberlabrat.core.Constants import REGION_CLASSES, COMPOUND_CLASSES


class Figure:

    def __new__(self, data_handler):

        @dataclass
        class Figure(Cacheable, data_handler):
            """
            Base class for all figures. Handles loading, saving, and plotting of figures.

            Attributes:
                project (str): The name of the project.
                compound (str|list, optional): The name of the compound. Defaults to None.
                region (str|list, optional): The name of the region. Defaults to None.
                experiment (str, optional): The name of the experiment. Defaults to None.
                p_value_threshold (float, optional): The p-value threshold used for statistical analysis defaults to value set at project level.
                handle_outliers (float, optional): Whether to handle outlier selection if any present. Defaults to True.
                remove_outliers (str, optional): Whether to remove outliers. Defaults to "eliminated".
                custom_params (dict, optional): Custom parameters for the figure. Defaults to an empty dictionary.
                extension (ClassVar[str]): The file extension for the figure. Defaults to "png".
            """

            custom_params: dict = field(kw_only=True, default_factory=dict)
            extension: ClassVar[str] = "png"

            def __post_init__(self):
                data_handler.__post_init__(self)
                self.setup()
                self.define_filename()
                Cacheable.__post_init__(self)

            def setup(self):
                self.compound_or_region = "compound" if self.is_compound() else "region"
                self.to_plot = "region" if self.is_compound() else "compound"
                self.order = COMPOUNDS_AND_REGIONS[self.to_plot].order(
                    self.data[self.to_plot].unique()
                )

            def is_compound(self):
                return isinstance(self.compound, str)

            def define_filename(self):
                if self.is_compound() == True:
                    if self.region in REGION_CLASSES.values():
                        self.filename = f"{self.compound} in {REGION_CLASSES.reversed.get(tuple(self.region), self.region)}"
                    else:
                        self.filename = f"{self.compound} in {self.region if self.region else 'all regions'}"
                else:
                    if self.compound in COMPOUND_CLASSES.values():
                        self.filename = f"{self.region} in {COMPOUND_CLASSES.reversed.get(tuple(self.compound), self.compound)}"
                    else:
                        self.filename = f"{self.region} in {self.compound if self.compound else 'all compounds'}"

            def save(self):
                self.fig.savefig(self.filepath)
                print(f"SAVED {self.filepath}")
                filepath_no_extension, _ = os.path.splitext(self.filepath)
                self.fig.savefig(f"{filepath_no_extension}.svg")
                print(f"SAVED {filepath_no_extension}.svg")

            def load(self):
                display(Image(filename=f"{self.filepath}"))

            def _repr_html_(self):
                if not hasattr(self, "fig"):
                    if self.is_saved:
                        self.load()
                    else:
                        raise Exception("Could not load figure")

        return Figure


@dataclass
class Histogram(Figure(QuantitativeDataSelection)):
    """
    Generate a histogram of treatments. If only one compound or region is specified, a simple histogram is generated.
    If multiple compounds or regions are specified, a summary histogram is generated.
    """

    plot_swarm: bool = field(default=True)
    figure_type: str = "histogram"

    def setup_plotter_parameters(self):
        multiple_regions = isinstance(self.region, list) or self.region is None
        multiple_compounds = isinstance(self.compound, list) or self.compound is None
        self.swarm_hue = self.hue = self.x = "treatment"

        self.palette = (
            self.experiment_information.palette
            if self.experiment
            else Palette(self.project).dict
        )
        self.significance_palette = {
            row.treatment: row.significance for row in Palette(self.project)
        }
        self.hue_order = self.treatments

        if self.pool == "treatment":
            self.x = self.to_plot
            self.hue = self.compound_or_region
            self.hue_order = self.order
            self.palette = self.significance_palette = None
            self.is_summary = multiple_compounds and multiple_regions
        else:
            self.is_summary = (
                isinstance(self.compound, list)
                or isinstance(self.region, list)
                or self.compound is None
                or self.region is None
            )
            
        self.title = (
            f"{self.compound or 'all compounds'} in {self.region or 'all regions'}"
        )
        self.ylabel = "" if "/" in self.compound else "ng/mg of tissue"
        if self.is_summary:
            self.ylabel = f"{self.compound_or_region} {self.ylabel} +/-98CI"
            self.title = ""

    def generate(self):
        self.setup_plotter_parameters()
        self.plot()
        if self.experiment:
            (
                self.label_summary_stats()
                if self.is_summary
                else self.label_histogram_stats()
            )

    def plot(self, custom_params=dict()):
        custom_params = {**self.custom_params, **custom_params}
        custom_params["palette"] = (
            {
                **self.palette,
                **self.custom_params.get("palette", {}),
                **custom_params.get("palette", {}),
            }
            if not self.pool == "treatment"
            else None
        )
        if self.is_summary:
            self.plot_summary(custom_params)
        else:
            self.plot_histogram(custom_params)

    def plot_histogram(self, custom_params):
        self.fig, self.ax = plt.subplots(figsize=(20, 10))
        ax = sns.barplot(
            data=custom_params.get("data", self.data),
            x=self.x,
            y="value",
            hue=self.hue,
            palette=custom_params.get("palette", self.palette),
            errorbar=custom_params.get(
                "errorbar", "sd"
            ),  # ("ci", 68) remi :) would be nice to also have SEM intergrated as an option - it is not inbuild like ci or sd in seabourn
            edgecolor=custom_params.get("edgecolor", ".2"),
            errcolor=custom_params.get("errcolor", ".2"),
            capsize=custom_params.get("capsize", 0.1),
            alpha=custom_params.get("alpha", 0.8),
            order=custom_params.get("hue_order", self.hue_order),
            dodge=custom_params.get("dodge", False),
        )
        if custom_params.get("plot_swarm", self.plot_swarm):
            ax = sns.swarmplot(
                data=custom_params.get("data", self.data),
                x=self.x,
                y="value",
                hue=custom_params.get("swarm_hue", self.swarm_hue),
                size=custom_params.get("size", 5),
                palette=custom_params.get(
                    "swarm_palette", custom_params.get("palette", self.palette)
                ),
                legend=custom_params.get("legend", False),
                order=custom_params.get("hue_order", self.hue_order),
                edgecolor=custom_params.get("edgecolor", "k"),
                linewidth=custom_params.get("linewidth", 1),
                linestyle=custom_params.get("linestyle", "-"),
                dodge=custom_params.get("dodge", False),
            )

        ax.tick_params(labelsize=custom_params.get("labelsize", 24))
        ax.set_ylabel(
            custom_params.get("ylabel", self.ylabel),
            fontsize=custom_params.get("ylabel_fontsize", 24),
        )
        ax.set_xlabel(
            " ", fontsize=custom_params.get("xlabel_fontsize", 20)
        )  # treatments
        ax.set_title(
            custom_params.get("title", self.title),
            y=custom_params.get("y", 1.04),
            fontsize=custom_params.get("fontsize", 34),
        )
        sns.despine(left=False)

    def plot_summary(self, custom_params):
        fig_width = custom_params.get("fig_width", 1.48 + 2 * len(self.order))
        self.fig, self.ax = plt.subplots(figsize=(fig_width, 10))
        self.ax = sns.barplot(
            data=self.data,
            x=self.to_plot,
            y="value",
            hue=self.hue,
            palette=custom_params.get("palette", self.palette),
            errorbar=custom_params.get("errorbar", "sd"),
            edgecolor=custom_params.get("edgecolor", ".2"),
            errcolor=custom_params.get("errcolor", ".2"),
            capsize=custom_params.get("capsize", 0.1),
            alpha=custom_params.get("alpha", 0.8),
            order=custom_params.get("order", self.order),  # self.order,
            hue_order=custom_params.get("hue_order", self.hue_order),
            errwidth=custom_params.get("errwidth", 1),
            dodge=custom_params.get("dodge", True),
            width=custom_params.get("bar_width", 0.8),
        )
        self.ax.tick_params(labelsize=40)
        self.ax.set_ylabel(
            self.ylabel, fontsize=44, labelpad=custom_params.get("labelpad", 100)
        )
        self.ax.yaxis.set_label_coords(
            custom_params.get("ylabel_x", -0.459 / fig_width), 0.5
        )
        self.ax.set_xlabel(" ", fontsize=20)  # remove x title
        self.ax.set_title(self.title, y=1.04, fontsize=34)
        self.ax.legend(loc="upper right")  # , bbox_to_anchor=(0.1, 1))
        self.ax.spines["top"].set_visible(False)
        self.ax.spines["right"].set_visible(False)
        plt.tight_layout()

    def label_histogram_stats(self):
        if self.statistic and self.statistic.is_significant:
            pairs, p_values = self.statistic.significant_pairs
            annotator = Annotator(
                self.ax,
                pairs,
                data=self.data,
                x="treatment",
                y="value",
                order=self.hue_order,
            )
            annotator.configure(text_format="star", loc="inside", fontsize="xx-large")
            annotator.set_pvalues_and_annotate(p_values)

    def label_summary_stats(self):
        for statistics in self.statistics:
            if statistics.is_significant:
                # Font Scaling # HARDCODE JJB TODO - also add significance pairs!
                base_font_size = 48
                scaling_factor = 0.2
                dynamic_font_size = max(
                    base_font_size - (scaling_factor * len(self.order)), 6
                )
                for pair in statistics.significant_pairs[0]:
                    for i, (treatment, symbol) in enumerate(
                        self.significance_palette.items()
                    ):
                        if treatment in pair:
                            hue = pair[
                                pair.index(treatment) - 1
                            ]  # work because only two elements 0 -> -1, 1 -> 0
                            x_index = self.order.index(
                                statistics.metadata[self.to_plot]
                            )  # Statistics metadata stores grouping info
                            hue_index = self.hue_order.index(hue)
                            bar = self.ax.patches[hue_index * len(self.order) + x_index]
                            self.ax.text(
                                bar.get_x() + bar.get_width() / 2,
                                (bar.get_height() * (1.3 + i / 5)),
                                symbol,
                                ha="center",
                                va="bottom",
                                fontsize=dynamic_font_size,
                            )
                            break

    def set(self, **kwargs):
        # kwargs["palette"] = {**self.palette, **kwargs.get("palette", {})}
        self.custom_params = kwargs
        self.initialize()


@dataclass
class MatricesFigure(Figure(DataSelection)):

    columns: list[str] = field(default=None)
    n_minimum: float = field(default=5)
    method: float = field(default="pearson")

    def __post_init__(self):
        if self.compound and "-" in self.compound:
            self.compound = self.compound.split("-")
        if self.region and "-" in self.region:
            self.region = self.region.split("-")
        super().__post_init__()

    def setup(self):
        super().setup()
        c_or_r = getattr(self, self.compound_or_region)
        self.var1 = c_or_r[0] if isinstance(c_or_r, list) else c_or_r
        self.var2 = c_or_r[-1] if isinstance(c_or_r, list) else c_or_r
        self.is_square = self.var1 != self.var2

    def is_compound(self):
        return super().is_compound() or len(self.compound) == 2

    def setup_plotter_parameters(self):
        self.build_matrices()
        self.homogenize_matrices()

    def build_matrices(self):
        cases = [
            Matrix(
                self.data.select(treatment=treatment),
                treatment,
                self.compound_or_region,
                self.var1,
                self.var2,
                self.to_plot,
                self.order,
                self.n_minimum,
                self.method,
                self.p_value_threshold,
            )
            for treatment in self.treatments
        ]  # Setup multiprocessing pool
        self.matrices = parallel_process(cases, description="Creating matrices")

    def homogenize_matrices(self):
        conserved_rows = set.intersection(
            *(set(matrix.corr_masked.index) for matrix in self.matrices)
        )
        conserved_cols = set.intersection(
            *(set(matrix.corr_masked.columns) for matrix in self.matrices)
        )

        for matrix in self.matrices:
            rows_to_drop = [
                row for row in matrix.corr_masked.index if row not in conserved_rows
            ]
            cols_to_drop = [
                col for col in matrix.corr_masked.columns if col not in conserved_cols
            ]
            matrix.corr_masked = matrix.corr_masked.drop(
                index=rows_to_drop, columns=cols_to_drop
            )

    def generate(self):
        self.setup_plotter_parameters()
        self.fig, self.axs = self.generate_figure()
        for i in range(len(self.axs)):
            self.plot_ax(i)

    def generate_figure(self):
        # determin number of treatments to corrispond to number of subplots
        num_treatments = len(self.matrices)
        num_cols = min(int(np.sqrt(num_treatments)), 2)  # max of 2 columns
        num_rows = (
            num_treatments + num_cols - 1
        ) // num_cols  # Compute the number of rows

        # define the base size and a scaling factor for the figure size
        base_size = 11
        scale_factor = 1

        # create subplots
        fig, axs = plt.subplots(
            num_rows,
            num_cols,
            figsize=(
                num_cols * base_size * scale_factor,
                num_rows * base_size * scale_factor,
            ),
            constrained_layout=True,
        )
        # fig.tight_layout(pad=2)
        # fig.subplots_adjust(hspace=0.4, wspace=0.4)
        return fig, [axs] if num_rows == num_cols == 1 else axs.flatten()

    def plot_ax(self, i):
        raise NotImplementedError("Must be implemented in subclass")


@dataclass
class Correlogram(MatricesFigure):

    figure_type: str = "correlogram"

    def plot_ax(self, i):
        ax = self.axs[i]
        matrix = self.matrices[i]
        title = f"{'->'.join([self.var1, self.var2]) if self.is_square else self.var1} in {matrix.grouping}"

        ax.set_title(
            title, fontsize=28, pad=20, y=1
        )  # Adjust the y position of the title manually for square correlogram

        sns.heatmap(
            matrix.corr_masked,
            vmin=-1,
            vmax=1,
            square=True,
            # annot=True, #R value annotations
            cmap="coolwarm",
            annot_kws={"size": 8},
            ax=ax,
            cbar_kws={"shrink": 0.7},  # adj color bar size
        )
        ax.set_xticklabels(
            ax.get_xticklabels(),
            rotation=45,
            ha="center",
            fontsize=12,
        )
        ax.set_yticklabels(
            ax.get_yticklabels(),
            rotation=45,
            va="center",
            fontsize=12,
        )

        ax.set_ylabel(matrix.var1, fontsize=28)
        ax.set_xlabel(matrix.var2, fontsize=28)


@dataclass
class Network(MatricesFigure):

    figure_type: str = "network"
    node_positions: dict = None  # Optional parameter for manual node positions

    def define_filename(self):
        super().define_filename()
        self.filename = self.filename.replace("-", "->")

    def setup_plotter_parameters(self):
        super().setup_plotter_parameters()
        self.networks = parallel_process(
            [NetworkModel(matrix) for matrix in self.matrices],
            description="Creating networks",
        )

    def generate(self):
        super().generate()
        fig, axs = self.generate_figure()
        for i in range(len(self.axs)):
            ax = axs[i]
            network = self.networks[i]
            self.plot_degrees(ax, network)

    def plot_ax(self, i):
        ax = self.axs[i]
        network = self.networks[i]
        title = f"{'->'.join([self.var1, self.var2]) if self.is_square else self.var1} in {network.matrix.grouping}"

        positions = self.node_positions if self.node_positions else network.pos

        nx.draw_networkx_nodes(
            network.G,
            positions,
            node_size=2000,
            alpha=0.95,
            node_color="white",
            edgecolors="black",
            ax=ax,
        )
        nx.draw_networkx_edges(
            network.G,
            positions,
            width=list(nx.get_edge_attributes(network.G, "weight").values()),
            edge_color=list(nx.get_edge_attributes(network.G, "color").values()),
            ax=ax,
            node_size=2000,
            **({"arrowstyle": "->", "arrowsize": 20} if network.is_directed else {}),
        )
        # Add labels to nodes
        node_labels = {
            node: node for node in network.G.nodes()
        }  # Label nodes with their names
        nx.draw_networkx_labels(
            network.G, positions, labels=node_labels, font_size=22, ax=ax
        )
        # nx.draw_networkx_edge_labels(
        #     network.G, positions, edge_labels=network.edge_labels, font_size=18, ax=ax
        # )

        # Set the aspect ratio to 'equal' for the plot area
        ax.set_aspect("equal")
        ax.margins(0.1)

        # Set title for the graph
        ax.set_frame_on(False)
        ax.set_title(title, fontsize=28, pad=-10, y=1)

    def plot_degrees(self, ax, network):
        """
        Plots histogram of node degrees from network with a standard distribution overlay
        input:
            network object
            ax to plot
        returns:
            ax to plot
        """
        title = f"{'->'.join([self.var1, self.var2]) if self.is_square else self.var1} in {network.matrix.grouping}"
        G = network.G  # Access the graph from the Network object
        all_nodes = list(G.nodes())
        degree_sequence = [d for n, d in G.degree()]
        node_labels_with_degrees = [(n, d) for n, d in G.degree()]

        mean_degree = np.mean(degree_sequence)
        std_degree = np.std(degree_sequence)

        # Use the max_node_degree property from the Network class
        max_degree = network.max_degree
        mean_degree = network.average_degree

        x = np.linspace(0, max_degree, 100)
        y = norm.pdf(x, mean_degree, std_degree)
        ax.plot(x, y, "r-", lw=2, label=f"Standard Distribution std={std_degree:.2f}")

        # Create the histogram
        counts, bins, patches = ax.hist(
            degree_sequence,
            bins=np.arange(max_degree + 2) - 0.5,
            edgecolor="black",
            alpha=0.8,
        )
        # Check if the sum of counts matches the number of nodes
        total_nodes = len(all_nodes)
        total_counted = sum(counts)
        if total_nodes != total_counted:
            raise ValueError(
                f"Total nodes ({total_nodes}) does not match total counted ({total_counted})"
            )

        # Annotate each bar with the corresponding node labels
        for i, patch in enumerate(patches):
            bin_center = patch.get_x() + patch.get_width() / 2
            labels = [n for n, d in node_labels_with_degrees if d == i]
            if labels:
                ax.text(
                    bin_center,
                    0.06 * patch.get_height(),
                    ", ".join(labels),
                    ha="center",
                    va="bottom",
                    fontsize=28,
                    rotation=90,
                )

        ax.set_title(
            title, fontsize=28, pad=20, y=1
        )  # Use the get_title property from Network class
        ax.set_xlabel("Degree", fontsize=22)
        ax.set_ylabel("Frequency (n nodes)", fontsize=22)
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.legend(fontsize=20)
        ax.tick_params(axis="x", labelsize=20)  # Adjust x-axis tick label size
        ax.tick_params(axis="y", labelsize=20)

        # HACKY PRINT
        title = network.get_title()
        print(title)
        print(
            f"edges = {network.total_edges}, pos = {network.pos_edges}, neg = {network.neg_edges}",
            f"density = {network.density}, max degree = {network.max_degree}, average degree = {network.average_degree}",
            f"unweighted clustering co = {network.avg_clust_coeff_unweighted}",
        )

        return ax


@dataclass
class Correlation(MatricesFigure):

    figure_type: str = "correlation"

    def define_filename(self):
        self.filename = f"{self.compound} in {self.region}"

    def generate(self):
        compound = self.compound if isinstance(self.compound, list) else [self.compound]
        region = self.region if isinstance(self.region, list) else [self.region]

        x_data = self.data.select(compound=compound[0], region=region[0])
        y_data = self.data.select(compound=compound[-1], region=region[-1])
        common_mouse_ids = list(set(x_data.mouse_id).intersection(set(y_data.mouse_id)))
        x_data = x_data.set_index("mouse_id").loc[common_mouse_ids].value.values
        y_data = y_data.set_index("mouse_id").loc[common_mouse_ids].value.values

        x_label = f"{compound[0]} in {region[0]} (ng/mg)"
        y_label = f"{compound[-1]} in {region[-1]} (ng/mg)"
        pearson_r, p_value = stats.pearsonr(x_data, y_data)
        color = "red" if pearson_r > 0 else "blue"

        # Create the plot
        self.fig, ax = plt.subplots()
        sns.scatterplot(x=x_data, y=y_data, ax=ax, marker="o", s=30, color="black")
        sns.regplot(
            x=x_data, y=y_data, ci=95, ax=ax, scatter=False, line_kws={"color": color}
        )

        ax.set_xlabel(x_label, fontsize=22)
        ax.set_ylabel(y_label, fontsize=22)
        ax.spines[["right", "top"]].set_visible(False)
        ax.set_title(self.treatment)

        # Add correlation values as text
        p_value_annotation = f"{p_value:.1e}" if p_value < 0.0001 else f"{p_value:.4f}"
        labels = f"Pearson R: {pearson_r:.2f}\np-value: {p_value_annotation}"
        ax.text(
            0.05,
            0.9,
            labels,
            transform=ax.transAxes,
            bbox=dict(facecolor="white", edgecolor="white", boxstyle="round"),
        )

        plt.tight_layout()
        plt.show()


@dataclass
class Table(ExcelDataset, Figure(QuantitativeDataSelection)):
    figure_type: str = field(default="table", init=False)

    def setup(self):
        self.order = REGIONS.order(self.data["region"].unique())

    def generate(self):
        grouped = (
            self.data.groupby(["region", "compound", "treatment"])
            .agg(
                mean_value=("value", "mean"),
                std_value=("value", lambda x: np.std(x, ddof=1)),
            )
            .reset_index()
        )

        # Combine mean and STD into a single string
        grouped["mean ± STD"] = grouped.apply(
            lambda row: f"{row['mean_value']:.3f} ± {row['std_value']:.3f}", axis=1
        )

        # Pivot the DataFrame
        pivot_df = grouped.pivot_table(
            index="region",
            columns=["compound", "treatment"],
            values="mean ± STD",
            aggfunc="first",
        )

        # Sort the multiindex columns
        return pivot_df.sort_index(axis=1).loc[self.order, self.compound]

    def load(self):
        return SelectableDataFrame(
            pd.read_excel(self.filepath, index_col=0, header=[0, 1])
        )


@dataclass
class StatisticsTable(Table):

    def define_filename(self):
        super().define_filename()
        self.filename += " STATS"

    def generate(self):

        
        if not self.statistics:
            return pd.DataFrame()
        
        results = []

        for statistic in self.statistics:
            data = statistic.results
            data = data[data["test"] == statistic.statistical_test][
                ["test", "region", "compound", "result_string"]
            ]
            results.append(data)
        results = pd.concat(results)
        
        results = results.pivot_table(
                index="region",
                columns=["test", "compound"],
                values="result_string",
                aggfunc="first",
            )
        
        results.index = pd.Categorical(results.index, categories=self.order, ordered=True)
        return results.sort_index()
        
        

    def load(self):
        return SelectableDataFrame(
            pd.read_excel(self.filepath, index_col=0, header=[0, 1])
        )



from setuptools import setup, find_packages

setup(
    name='cyberlabrat',
    version='0.1.9',
    description='Contains the classes used for generix data analysis',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Jasmine Butler & Remi Corne',
    author_email='remi.corne@gmail.com',
    url='https://github.com/jjb-hub/phd',
    packages=find_packages(),
    install_requires=[
        "pandas==2.0.3",
        "matplotlib==3.7.3",
        "outlier_utils",
        "pingouin",
        "scipy",
        "remi_statannotations_fork",
        "setuptools",
        "statsmodels",
        "openpyxl",
        "pydantic",
        "pytest",
        "tqdm",
        "ipykernel",
        "ipywidgets",
        "IPython",
        "networkx",
        ], 
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',
)




import os


class FileSystem:
    """
    Static class that handles the filesystem architecture.
    Automatically creates folders if they don't exist.
    Automatically extracts the project, experiment and figure_type from the path.

    """

    ROOT = os.path.dirname(os.path.abspath(__file__))
    PROJECTS = os.path.join(ROOT, "PROJECTS")
    CONSTANTS = os.path.join(ROOT, "json")

    PATH_ELEMENT_ORDER = ["project", "experiment", "figure_type"]
    
    @staticmethod
    def list_projects():
        projects = os.listdir(FileSystem.PROJECTS)
        return [project.split("/")[-1] for project in projects]

    @staticmethod
    def list_experiments(project=None):
        if project:
            experiments = []
            for file in os.listdir(project):
                if os.path.isdir(file):
                    experiments.append(file.split("/")[-1])
            return experiments
        else:
            experiments = {}
            for project in FileSystem.list_projects():
                experiments[project] = FileSystem.list_experiments(project)
            return experiments

    @staticmethod
    def get_location(**path_elements):
        location = FileSystem.PROJECTS
        for element in FileSystem.PATH_ELEMENT_ORDER:
            if path_elements.get(element):
                location = os.path.join(location, path_elements[element])
                if not os.path.exists(location):
                    os.mkdir(location)
                    print(f"Created {path_elements[element]} {element} folder")
        return location



import os, re
from dataclasses import dataclass
from typing import ClassVar
import pandas as pd
import numpy as np
from cyberlabrat.core.Dataset import PickleDataset, SelectableDataFrame
from cyberlabrat.core.Metadata import (
    ProjectInformation,
    ExperimentInformation,
    TreatmentInformation,
)
from cyberlabrat.core.questions import yes_or_no, input_escape
from cyberlabrat.core.Constants import REGIONS, COMPOUNDS
from tqdm import tqdm
from outliers import smirnov_grubbs as grubbs
from cyberlabrat.core.utils import parallel_process


def detect_raw_data(project):
    for filename in os.listdir():
        if project.name.upper() in filename.upper():
            if os.path.splitext(filename)[1].lower() in [".csv", ".xlsx"]:
                is_correct = yes_or_no(f"Detected {filename}. Confirm?")
                if is_correct:
                    return filename
    return None


def handle_raw_col_name(column):
    match = re.match(r"(\w+)[-_ ](\w+)", column)
    while not match or not len(match.groups()) == 2:
        column = input(f"Wrong format: '{column}', input new 'compound_region': ")
        match = re.match(r"(\w+)[-_ ](\w+)", column)
    return match.groups()


@dataclass(repr=False)
class RawHPLC(PickleDataset):

    project: str
    filename: ClassVar[str] = "raw_hplc"

    def generate(self):
        project_information = ProjectInformation(self.project)
        filepath = f"{os.getcwd()}/{project_information.raw_data_filename}"
        extension = os.path.splitext(project_information.raw_data_filename)[1].lower()
        raw_data = (
            pd.read_excel(filepath) if extension == ".xlsx" else pd.read_csv(filepath)
        )
        return raw_data.rename(columns=self.get_valid_columns(raw_data.columns))

    def get_valid_columns(self, columns):
        mandatory_columns = ["mouse_id", "group_id"]
        absent_columns = [col for col in mandatory_columns if col not in columns]
        if absent_columns:
            raise ValueError(f"Absent columns: {(', ').join(absent_columns)}")
        compound_region_tuples = [
            handle_raw_col_name(col) for col in columns if col not in mandatory_columns
        ]
        invalid_compounds = {
            compound
            for compound, _ in compound_region_tuples
            if compound not in COMPOUNDS
        }
        if invalid_compounds:
            print(f"Invalid compounds: {invalid_compounds}")

        invalid_regions = {
            region for _, region in compound_region_tuples if region not in REGIONS
        }
        if invalid_regions:
            print(f"Invalid regions: {invalid_regions}")

        compound_translator = {
            invalid_compound: COMPOUNDS.get_valid_choice(invalid_compound)
            for invalid_compound in invalid_compounds
        }
        region_translator = {
            invalid_region: REGIONS.get_valid_choice(invalid_region)
            for invalid_region in invalid_regions
        }
        return {
            f"{compound}_{region}": f"{compound if compound in COMPOUNDS else compound_translator[compound]}_{region if region in REGIONS else region_translator[region]}"
            for compound, region in compound_region_tuples
        }


class ProjectSelectableDataframe(SelectableDataFrame):
    """
    Shameful hack.
    TODO: Eliminate asap
    """
    
    def __init__(self, data=None, project=None, *args, **kwargs):
        super().__init__(data, *args, **kwargs)
        self.project = project
        
    @property
    def _constructor(self):
        return lambda *args, **kwargs: ProjectSelectableDataframe(*args, project=self.project, **kwargs) if hasattr(self, "project") else SelectableDataFrame(*args, **kwargs)

    def select(self, **selector) -> SelectableDataFrame:
        experiment = selector.pop("experiment", None)
        if experiment:
            experiment = ExperimentInformation(self.project).select(label=experiment)
            selector["group_id"] = experiment.groups
        data = SelectableDataFrame(self).select(**selector)
        return ProjectSelectableDataframe(data, project=self.project) if hasattr(self, "project") else data

@dataclass(repr=False)
class HPLC(PickleDataset):

    project: str
    filename: ClassVar[str] = "hplc"

    def generate(self):
        raw_data = RawHPLC(self.project).df
        compound_data = raw_data.melt(
            id_vars=["mouse_id", "group_id"], value_vars=raw_data.columns[2:]
        )
        compound_data[["compound", "region"]] = compound_data["variable"].str.split(
            "_", expand=True
        )
        compound_data = compound_data.drop(columns=["variable"])
        ratio_data = pd.merge(
            left=compound_data,
            right=compound_data,
            on=[
                "mouse_id",
                "group_id",
                "region",
            ],
            suffixes=["_1", "_2"],
        ).reset_index(
            drop=True
        )  # merge every compound to every other for each mouse, we want to reset the index (ie make sure it has no duplicates) becaus many update operations will use it

        ratio_data = ratio_data[(ratio_data.compound_1 != ratio_data.compound_2)]

        def calculateRatio(row):
            ratio_name = f"{row.compound_1}/{row.compound_2}"
            return [
                ratio_name,
                row.value_1 / row.value_2 if row.value_1 and row.value_2 else np.nan,
            ]

        tqdm.pandas(desc="Calculating ratios", unit="ratio")

        ratio_data[["compound", "value"]] = ratio_data.progress_apply(
            calculateRatio,
            axis=1,
            result_type="expand",
        )  # calculate ratio
        # Drop duplicate columns
        compound_and_ratios_df = pd.concat(
            [
                compound_data,
                ratio_data.drop(
                    columns=["compound_1", "compound_2", "value_1", "value_2"]
                ),
            ]
        )
        return compound_and_ratios_df.replace(0, np.nan)
    
    def select(self, **selector) -> ProjectSelectableDataframe:
        return self.full_df.select(**selector)

    @property
    def full_df(self) -> ProjectSelectableDataframe:
        data = ProjectSelectableDataframe(pd.concat([self.df, TissueWeight(self.project).df]))
        data = ProjectSelectableDataframe(
            data.extend(Outliers(self.project)).extend(
                TreatmentInformation(self.project)
            ), self.project
        )
        return data[[col for col in data.columns if "Unnamed" not in col]]

    @property
    def compounds(self):
        return self.df.compound.unique()

    @property
    def regions(self):
        return self.df.region.unique()

    @property
    def compounds_and_regions(self):
        return {
            "compounds": self.df.compound.unique(),
            "regions": self.df.region.unique(),
        }


def grubbs_test(values, p_value_threshold):
    """
    Takes a list of values on which to perform the test and returns normal values
    """
    return grubbs.test(values, alpha=float(p_value_threshold))


def get_labeled_df(df__test__p_value_threshold):
    # if standar variation is 0, we can't calculate outliers
    df, test, p_value_threshold = df__test__p_value_threshold
    only_values = df[df.value != 0].dropna()
    if only_values.value.count() < 3:
        df["outlier_status"] = False
        return df
    outlier_test = OUTLIER_TESTS[test]
    normal_values = outlier_test(only_values.value.tolist(), p_value_threshold)
    df["is_outlier"] = df.value.apply(lambda value: value not in normal_values)
    df["outlier_status"] = df.is_outlier.apply(
        lambda is_outlier: "suspected" if is_outlier else "normal"
    )
    return df


OUTLIER_TESTS = {"grubbs": grubbs_test}


@dataclass(repr=False)
class Outliers(PickleDataset):

    project: str
    filename: ClassVar[str] = "outliers"

    def generate(self):
        hplc = HPLC(self.project)
        project_information = ProjectInformation(self.project)
        cases = [
            (
                subset_df,
                project_information.outlier_test,
                project_information.p_value_threshold,
            )
            for _, subset_df in hplc.df.groupby(
                ["group_id", "compound", "region", "region"]
            )
        ]
        results = parallel_process(
            cases, get_labeled_df, description="Calculating outliers"
        )
        return pd.concat(results)

    def update(self, updates):
        data = self.df.set_index(["compound", "region", "mouse_id", "group_id"])
        updates = updates.set_index(["compound", "region", "mouse_id", "group_id"])
        data.update(updates[["outlier_status"]])
        self.save(data.reset_index())

   
@dataclass 
class TissueWeight(PickleDataset):
    
    project: str
    filename: ClassVar[str] = "tissue_weight"
    
    def generate(self):
        filename = None
        while not os.path.exists(f"{os.getcwd()}/{filename}"):
            filename = input_escape("Enter tissue weight filename: ")
        filepath = f"{os.getcwd()}/{filename}"
        extension = os.path.splitext(filename)[1].lower()
        raw_data = pd.read_excel(filepath) if extension == ".xlsx" else pd.read_csv(filepath)
        raw_data.rename(columns={compound: REGIONS.get_valid_choice(compound) for compound in raw_data.columns[1:]})
        raw_data = Groups(self.project).extend(raw_data)
        raw_data = raw_data.melt(
            id_vars=["mouse_id", "group_id"], value_vars=raw_data.columns[2:]
        )
        raw_data["compound"] = "weight"
        return raw_data.rename(columns={"variable": "region"})
        

@dataclass
class Groups(PickleDataset):
    
    project: str
    filename: ClassVar[str] = "groups"
    
    def generate(self):
        hplc_data = HPLC(self.project).df[["mouse_id", "group_id"]]
        return hplc_data.drop_duplicates()